/* Copyright (C) RSA Data Security, Inc. created 1995, 1998.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

/*  This program illustrates the process of calling B_DecryptUpdate multiple
      times.  The data to be decrypted is in the file "encrfile.dat".  It will
      read a specified number of bytes from the file, decrypt it, then write
      the decrypted bytes out to an output file, "decrfile.dat", until the
      end of file is reached.
 */

#include <stdio.h>
#include <stdlib.h>       
#include <string.h>      
#include "aglobal.h"
#include "bsafe.h"
#include "demochos.h"

#define UPDATE_SIZE        64
#define UPDATE_OUTPUT_SIZE (UPDATE_SIZE + 16)

int GetDataFromFile PROTO_LIST
  ((FILE *, unsigned int, unsigned char *, unsigned int *, int *));

int AppendDataToFile PROTO_LIST
  ((FILE *, unsigned char *, unsigned int));
                     
void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));

void main()
{
  B_KEY_OBJ rc4Key = (B_KEY_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ decryptionObject = (B_ALGORITHM_OBJ)NULL_PTR;

  static unsigned char rc4KeyData[10] = {
    0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12
  };
  unsigned int rc4KeyDataLen = 10;
  ITEM rc4KeyItem;
  
  FILE *inputFile = (FILE *)NULL_PTR;
  FILE *outputFile = (FILE *)NULL_PTR;

  unsigned char dataToDecrypt[UPDATE_SIZE];
  unsigned char blockOfDecryptedData[UPDATE_OUTPUT_SIZE];
  unsigned int dataToDecryptLen, totalBytesSoFar;
  unsigned int outputLenUpdate, outputLenFinal;
  unsigned int sizeToUpdate = UPDATE_SIZE;
  int endFlag, status;

  inputFile = fopen("encrfile", "rb");
  if ((status = (inputFile == (FILE *)NULL_PTR)) != 0) {
    printf ("\n Usage: multdecr < encrfile\n");
    printf ("\n(\"encrfile\" is the output file created by running");
    printf (" \"multencr.exe\" first.)\n");
    exit(0);
  }
       
  do {
    printf ("RC4 algorithm - Multiple Updates \n");
    printf ("================================ \n");
    printf ("\n");

    printf ("... Decryption phase (calling Update multiple times) \n");
    printf ("================================================== \n");

    outputFile = fopen("decrfile", "wb");
    if ((status = (outputFile == (FILE *)NULL_PTR)) != 0)
      break;
    
    totalBytesSoFar = 0;

    /*  Create a key object. */
    if ((status = B_CreateKeyObject (&rc4Key)) != 0)
      break;

    /*  Set the key to be a RC4 key and set it to the rc4KeyData declared
        above. */
    rc4KeyItem.data = rc4KeyData;
    rc4KeyItem.len = rc4KeyDataLen;

    if ((status = B_SetKeyInfo
         (rc4Key, KI_Item, (POINTER)&rc4KeyItem)) != 0)
      break;

    /*  Create an algorithm object. */
    if ((status = B_CreateAlgorithmObject
         (&decryptionObject)) != 0)
      break;

    /*  Set the algorithm to a type that does RC4 encryption */
    if ((status = B_SetAlgorithmInfo
         (decryptionObject, AI_RC4, NULL_PTR)) != 0)
      break;
    
    /*  Ready to decrypt. Use B_DecryptInit, B_DecryptUpdate
          and B_DecryptFinal
     */
    if ((status = B_DecryptInit
         (decryptionObject, rc4Key, DEMO_ALGORITHM_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    while ((status = GetDataFromFile
            (inputFile, sizeToUpdate, dataToDecrypt,
             &dataToDecryptLen, &endFlag)) == 0) {
      printf ("dataToDecryptLen = %i \n", dataToDecryptLen);
      PrintBuf (dataToDecrypt, dataToDecryptLen);
      if ((status = B_DecryptUpdate
           (decryptionObject, blockOfDecryptedData,
            &outputLenUpdate, UPDATE_OUTPUT_SIZE, dataToDecrypt,
            dataToDecryptLen, (B_ALGORITHM_OBJ)NULL_PTR,
            (A_SURRENDER_CTX *)NULL_PTR)) != 0)
        break;

      /*  Save the decrypted data  */
      if ((status = AppendDataToFile
           (outputFile, blockOfDecryptedData,
            outputLenUpdate)) != 0)
        break;

      totalBytesSoFar += outputLenUpdate;
      if (endFlag == 1)
        break;
    } /*  end while  */

    /*  If there was an error in the above while loop, break out of the
          do-while construct.  */
    if (status != 0)
      break;
    
    /*  Call B_decryptFinal once after all Updates.  */           
    if ((status = B_DecryptFinal
         (decryptionObject, blockOfDecryptedData, &outputLenFinal,
          UPDATE_OUTPUT_SIZE, (B_ALGORITHM_OBJ)NULL_PTR,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Save the decrypted data  */
    if ((status = AppendDataToFile
         (outputFile, blockOfDecryptedData,
          outputLenFinal)) != 0)
      break;

    totalBytesSoFar += outputLenFinal;
    
    printf ("  Total length of decrypted data = %i ", totalBytesSoFar);

  } while (0);

  if (status != 0) {
    printf ("Status = %i \n", status);
    printf ("Multiple Updating failed");
  }
  
  /*  Done with the key and algorithm objects, so rc4troy them.
   */
  B_DestroyKeyObject (&rc4Key);
  B_DestroyAlgorithmObject (&decryptionObject);
  
  /*  Free up any memory allocated, save it to a file or print it out first
      if you need to save it.
   */

  T_memset (dataToDecrypt, 0, sizeof (dataToDecrypt));

  if (inputFile != (FILE *)NULL_PTR)
    fclose (inputFile);
  if (outputFile != (FILE *)NULL_PTR)
    fclose (outputFile);

} /*  end main  */


int GetDataFromFile
  (inputFile, sizeToUpdate, dataToDecrypt, dataToDecryptLen, endFlag)
FILE *inputFile;
unsigned int sizeToUpdate;
unsigned char *dataToDecrypt;
unsigned int *dataToDecryptLen;
int *endFlag;
{
  unsigned char dummy;

  do {    
    *dataToDecryptLen = fread (dataToDecrypt, 1, sizeToUpdate, inputFile);
    if (*dataToDecryptLen == sizeToUpdate)
    /* Read exactly sizeToUpdate bytes, so reading one more will set 
         end of file if there were exactly sizeToUpdate bytes in the file.
     */
      fread (&dummy, 1, 1, inputFile);
  
    if (feof (inputFile)) {
      *endFlag = 1;
      fclose (inputFile);
    }
    else {
      *endFlag = 0;
      fseek (inputFile, -1, SEEK_CUR);
    }
  } while (0);    
  return (0);
}  /*  end GetDataFromFile  */


int AppendDataToFile
  (outputFile, blockOfDecryptedData, outputLenUpdate)
FILE *outputFile;
unsigned char *blockOfDecryptedData;
unsigned int outputLenUpdate;
{
  do {  
    if (fwrite (blockOfDecryptedData, 1, outputLenUpdate, outputFile)
          < outputLenUpdate) {
      fclose (outputFile);
    }
  } while (0);
  return (0);
}  /*  end AppendDataToFile  */


/*  This procedure will print out what's in the buffer.
 */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */
    


