/* Copyright (C) RSA Data Security, Inc. created 1995, 1998.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

/*  This program illustrates the process of calling B_EncryptUpdate multiple
    times.  The data to be encrypted is in the file "infile.dat".  It will
    read a specified number of bytes from the file, encrypt it, then write
    the encrypted bytes out to an output file, "encrfile.dat", until the
    end of file is reached.
    To get back the original data or plaintext, compile "multdecr.c" and
    run.
 */

#include <stdio.h>
#include <stdlib.h>       
#include <string.h>      
#include "aglobal.h"
#include "bsafe.h"
#include "demochos.h"

#define UPDATE_SIZE        64
#define UPDATE_OUTPUT_SIZE (UPDATE_SIZE + 16)

int GetDataFromFile PROTO_LIST
  ((FILE *, unsigned int, unsigned char *, unsigned int *, int *));

int AppendDataToFile PROTO_LIST
  ((FILE *, unsigned char *, unsigned int));
                     
void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));

void main()
{
  B_KEY_OBJ rc4Key = (B_KEY_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ encryptionObject = (B_ALGORITHM_OBJ)NULL_PTR;

  static unsigned char rc4KeyData[10] = {
    0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x10, 0x11, 0x12
  };
  unsigned int rc4KeyDataLen = 10;
  ITEM rc4KeyItem;
  
  FILE *inputFile = (FILE *)NULL_PTR;
  FILE *outputFile = (FILE *)NULL_PTR;

  unsigned char dataToEncrypt[UPDATE_SIZE];
  unsigned char blockOfEncryptedData[UPDATE_OUTPUT_SIZE];
  unsigned int dataToEncryptLen, totalBytesSoFar;
  unsigned int outputLenUpdate, outputLenFinal;
  unsigned int sizeToUpdate = UPDATE_SIZE;
  int endFlag, status;

  inputFile = fopen("infile", "rb");

  if ((status = (inputFile == (FILE *)NULL_PTR)) != 0) {
    printf ("\n Usage:  multencr < infile \n");
    printf ("\n (\"infile\" is a hard-coded input file.");
    printf ("  Please create one as needed.)\n");	    
    exit(0);
  };  
  
  do {
    printf ("RC4 algorithm - Multiple Updates \n");
    printf ("============================================== \n");
    printf ("\n");

    printf ("... Encryption phase (calling Update multiple times) \n");
    printf ("================================================== \n");

    outputFile = fopen("encrfile", "wb");
    if ((status = (outputFile == (FILE *)NULL_PTR)) != 0)
      break;
    
    totalBytesSoFar = 0;

    /*  Create a key object. */
    if ((status = B_CreateKeyObject (&rc4Key)) != 0)
      break;
    
    /*  Set the key to be a RC4 key and set it to the rc4KeyData declared
        above. */
    rc4KeyItem.data = rc4KeyData;
    rc4KeyItem.len = rc4KeyDataLen;

    if ((status = B_SetKeyInfo
         (rc4Key, KI_Item, (POINTER)&rc4KeyItem)) != 0)
      break;

    /*  Create an algorithm object. */
    if ((status = B_CreateAlgorithmObject
         (&encryptionObject)) != 0)
      break;

    /*  Set the algorithm to a type that does DES encryption in CBC mode.
        This algorithm type needs an 8-byte initialization vector, use
        initVector declared above.  */
    if ((status = B_SetAlgorithmInfo
         (encryptionObject, AI_RC4, NULL_PTR)) != 0)
      break;

    /*  Ready to encrypt. Use B_EncryptInit, B_EncryptUpdate
          and B_EncryptFinal
     */
    if ((status = B_EncryptInit
         (encryptionObject, rc4Key, DEMO_ALGORITHM_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    while ((status = GetDataFromFile
            (inputFile, sizeToUpdate, dataToEncrypt,
             &dataToEncryptLen, &endFlag)) == 0) {
      printf ("dataToEncryptLen = %i \n", dataToEncryptLen);
      PrintBuf (dataToEncrypt, dataToEncryptLen);
      if ((status = B_EncryptUpdate
           (encryptionObject, blockOfEncryptedData,
            &outputLenUpdate, UPDATE_OUTPUT_SIZE, dataToEncrypt,
            dataToEncryptLen, (B_ALGORITHM_OBJ)NULL_PTR,
            (A_SURRENDER_CTX *)NULL_PTR)) != 0)
        break;

      /*  Save the encrypted data  */
      if ((status = AppendDataToFile
           (outputFile, blockOfEncryptedData,
            outputLenUpdate)) != 0)
        break;

      totalBytesSoFar += outputLenUpdate;
      if (endFlag == 1)
        break;
    } /*  end while  */

    /*  If there was an error in the above while loop, break out of the
          do-while construct.  */
    if (status != 0)
      break;
    
    /*  Call B_EncryptFinal once after all Updates.  */           
    if ((status = B_EncryptFinal
         (encryptionObject, blockOfEncryptedData, &outputLenFinal,
          UPDATE_OUTPUT_SIZE, (B_ALGORITHM_OBJ)NULL_PTR,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Save the encrypted data  */
    if ((status = AppendDataToFile
         (outputFile, blockOfEncryptedData,
          outputLenFinal)) != 0)
      break;

    totalBytesSoFar += outputLenFinal;
    
    printf ("  Total length of encrypted data = %i ", totalBytesSoFar);

  } while (0);

  if (status != 0) {
    printf ("Status = %i \n", status);
    printf ("Multiple Updating failed");
  }
  
  /*  Done with the key and algorithm objects, so destroy them.
   */
  B_DestroyKeyObject (&rc4Key);
  B_DestroyAlgorithmObject (&encryptionObject);
  
  /*  Free up any memory allocated, save it to a file or print it out first
      if you need to save it.
   */

  T_memset (dataToEncrypt, 0, sizeof (dataToEncrypt));

  if (inputFile != (FILE *)NULL_PTR)
    fclose (inputFile);
  if (outputFile != (FILE *)NULL_PTR)
    fclose (outputFile);

} /*  end main  */


int GetDataFromFile
  (inputFile, sizeToUpdate, dataToEncrypt, dataToEncryptLen, endFlag)
FILE *inputFile;
unsigned int sizeToUpdate;
unsigned char *dataToEncrypt;
unsigned int *dataToEncryptLen;
int *endFlag;
{
  unsigned char dummy;

  do {    
    *dataToEncryptLen = fread (dataToEncrypt, 1, sizeToUpdate, inputFile);
    if (*dataToEncryptLen == sizeToUpdate)
    /* Read exactly sizeToUpdate bytes, so reading one more will set 
         end of file if there were exactly sizeToUpdate bytes in the file.
     */
      fread (&dummy, 1, 1, inputFile);
  
    if (feof (inputFile)) {
      *endFlag = 1;
      fclose (inputFile);
    }
    else {
      *endFlag = 0;
      fseek (inputFile, -1, SEEK_CUR);
    }
  } while (0);    
  return (0);
}  /*  end GetDataFromFile  */


int AppendDataToFile
  (outputFile, blockOfEncryptedData, outputLenUpdate)
FILE *outputFile;
unsigned char *blockOfEncryptedData;
unsigned int outputLenUpdate;
{
  do {  
    if (fwrite (blockOfEncryptedData, 1, outputLenUpdate, outputFile)
          < outputLenUpdate) {
      fclose (outputFile);
    }
  } while (0);
  return (0);
}  /*  end AppendDataToFile  */


/*  This procedure will print out what's in the buffer.
 */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */
    


