/* Copyright (C) RSA Data Security, Inc. created 1994.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

#include <stdio.h>
#include <string.h>
#include "aglobal.h"
#include "bsafe.h"
#include "demochos.h"
#include "bsliteds.h"

/* If compiling with Borland, define BORLAND_STACK to the stack size,
     such as -DBORLAND_STACK=15000
 */
#ifdef BORLAND_STACK
extern unsigned _stklen = BORLAND_STACK;
#endif

/* If compiling with non-C calling format, define MAIN_CALL to force
     the compiler back to C calling format for main.  Otherwise, this
     sets MAIN_CALL to nothing.
 */
#ifndef MAIN_CALL
#define MAIN_CALL
#endif

int MAIN_CALL main PROTO_LIST ((int, char **));
static int SetOptions PROTO_LIST ((int, char **));
static int DoGenerateKeys PROTO_LIST ((void));
static void DoSignFile PROTO_LIST ((void));
static void DoVerifyFile PROTO_LIST ((void));
static void PrintMessage PROTO_LIST ((char *));
static void PrintError PROTO_LIST ((char *, int));
static int GetCommand PROTO_LIST ((char *, unsigned int, char *));
static char *BSAFE2_ErrorString PROTO_LIST ((int));

static int SILENT_PROMPT = 0;

static unsigned char PUBLIC_KEY_DATA[512];
static unsigned int PUBLIC_KEY_LEN = 0;
static unsigned char PRIVATE_KEY_DATA[1024];
static unsigned int PRIVATE_KEY_LEN = 0;
static unsigned char SEED[256];

int MAIN_CALL main (argc, argv)
int argc;
char *argv[];
{
  char command[100];
  int done = 0;

  if (SetOptions (argc, argv))
    return (0);
  
  /* Print version message.
   */
  strcpy (command, "Command Line DSA Demo for BSAFE version ");
  strcpy (command + strlen (command), BSAFE_VERSION);
  PrintMessage (command);
  if (DoGenerateKeys () != 0)
    return (-1);

  while (!done) {
    PrintMessage ("");
    PrintMessage ("S - Sign a file using DSA/SHA");
    PrintMessage ("V - Verify a DSA signed file");
    PrintMessage ("Q - Quit");
    GetCommand (command, sizeof (command), "  Enter choice: ");
    
    switch (*command) {
    case '\0':
    case '#':
      /* entered a blank line or a comment */
      break;
      
    case 's':
    case 'S':
      DoSignFile ();
      break;
      
    case 'v':
    case 'V':
      DoVerifyFile ();
      break;
      
    case 'Q':
    case 'q':
      done = 1;
      break;
      
    default:
      PrintError ("ERROR: Unrecognized command.  Try again.", 0);
      break;
    }
  }
  
  return (0);
}

/* Set options from command line and return 0 for success, 1 for bad format.
 */
static int SetOptions (argc, argv)
int argc;
char *argv[];
{
  int i, status = 0;
  
  for (i = 1; i < argc; i++) {
    if (argv[i][0] != '-') {
      status = 1;
      break;
    }
    
    if (argv[i][1] == 's')
      SILENT_PROMPT = 1;
    else {
      status = 1;
      break;
    }
  }

  if (status)
    puts ("Usage: bdemo [-s]\n  -s silent prompts");

  return (status);
}

static void DoSignFile ()
{
  char inputFilename[256], signatureFilename[256];
  int status;
  if (GetCommand (inputFilename, sizeof (inputFilename),
              "  Enter filename of content to sign (blank to cancel): "))
    return;
  if (GetCommand (signatureFilename, sizeof (signatureFilename),
              "  Enter filename to save the signature (blank to cancel): "))
    return;

  if ((*inputFilename == '\0') || (*signatureFilename == '\0'))
    return;
  
  if ((status = BSL_SHAWithDSASignFile
       (signatureFilename, inputFilename, PRIVATE_KEY_DATA, PRIVATE_KEY_LEN,
        (unsigned char *)NULL_PTR, 0, (unsigned char *)NULL_PTR, 0,
        SEED, strlen ((char *) SEED))) != 0)
    PrintError ("signing file", status);
  else
    PrintMessage ("File Signed.");
}

static void DoVerifyFile ()
{
  char inputFilename[256], signatureFilename[256];
  int status;
  
  if (GetCommand (inputFilename, sizeof (inputFilename),
              "  Enter name of file to verify (blank to cancel): "))
    return;
  if (GetCommand (signatureFilename, sizeof (signatureFilename),
              "  Enter filename of signature (blank to cancel): "))
    return;

  if ((*inputFilename == '\0') || (*signatureFilename == '\0'))
    return;
  
  if ((status = BSL_SHAWithDSAVerifyFile
       (signatureFilename, inputFilename, PUBLIC_KEY_DATA,
        PUBLIC_KEY_LEN)) != 0) {
    PrintError ("verifying file", status);
    return;
  }
  else
    PrintMessage ("Signature verified.");
}

static int DoGenerateKeys ()
{
  int status;

  if ((GetCommand ((char *)SEED, sizeof (SEED),
       "  Enter seed to generate the DSA keypair (blank to cancel): ")) != 0)
    return (1);

  PrintError ("  Generating DSA keypair, please wait...", 0);

  if ((status = BSL_MakeDSAKeyPairFromSeed
       (PUBLIC_KEY_DATA, &PUBLIC_KEY_LEN, PRIVATE_KEY_DATA,
        &PRIVATE_KEY_LEN, 1024, (unsigned char *)NULL_PTR, 0,
        512, (unsigned char *)SEED,
        (unsigned int)strlen ((char *)SEED),
        (unsigned char *)NULL_PTR, 0)) != 0) {
    PrintError ("generating DSA keys", status);
    return (1);
  }

  PrintMessage ("DSA public key and private key are now ready to use.");
  return (0);
}

static void PrintMessage (message)
char *message;
{
  if (!SILENT_PROMPT) {
    puts (message);
    fflush (stdout);
  }
}

/* If type is zero, simply print the task string, otherwise convert the
     type to a string and print task and type.
 */
static void PrintError (task, type)
char *task;
int type;
{
  char *typeString, buf[80];

  if (type == 0) {
    puts (task);
    return;
  }
  
  /* Convert the type to a string if it is recognized.
   */
  if ((typeString = BSAFE2_ErrorString (type)) == (char *)NULL_PTR) {
    sprintf (buf, "Code 0x%04x", type);
    typeString = buf;
  }

  printf ("ERROR: %s while %s\n", typeString, task);  
  fflush (stdout);
}

static int GetCommand (command, maxCommandSize, prompt)
char *command;
unsigned int maxCommandSize;
char *prompt;
{
  unsigned int i;
  
  if (!SILENT_PROMPT) {
    printf ("%s\n", prompt);  
    fflush (stdout);
  }

  fgets (command, maxCommandSize, stdin);

  if (*command == '\n' || *command == '\r')
    return (-1);
  
  /* Replace the line terminator with a '\0'.
   */
  for (i = 0; command[i] != '\0'; i++) {
    if (command[i] == '\n' || command[i] == '\r' ||
        i == (maxCommandSize - 1)) {
      command[i] = '\0';
      return (0);
    }
  }
  return (0);
}

/* Return the error string for the type, or (char *)NULL_PTR if the type
     is not recognized.
 */
static char *BSAFE2_ErrorString (type)
int (type);
{
  switch (type) {
  case BE_ALGORITHM_ALREADY_SET:
    return ("Algorithm object has already been set with algorithm info");
  case BE_ALGORITHM_INFO:
    return ("Invalid algorithm info format");
  case BE_ALGORITHM_NOT_INITIALIZED:
    return ("Algorithm object has not been initialized");
  case BE_ALGORITHM_NOT_SET:
    return ("Algorithm object has not been set with algorithm info");
  case BE_ALGORITHM_OBJ:
    return ("Invalid algorithm object");
  case BE_ALG_OPERATION_UNKNOWN:
    return ("Unknown operation for an algorithm or algorithm info type");
  case BE_ALLOC:
    return ("Insufficient memory");
  case BE_CANCEL:
    return ("Operation was cancelled by the surrender function");
  case BE_DATA:
    return ("Generic data error");
  case BE_EXPONENT_EVEN:
    return ("Invalid even value for public exponent in keypair generation");
  case BE_EXPONENT_LEN:
    return
      ("Invalid exponent length for public exponent in keypair generation");
  case BE_HARDWARE:
    return ("Cryptographic hardware error");
  case BE_INPUT_DATA:
    return ("Invalid format for input data");
  case BE_INPUT_LEN:
    return ("Invalid length for input data");
  case BE_KEY_ALREADY_SET:
    return ("Key object has already been set with key info");
  case BE_KEY_INFO:
    return ("Invalid key info format");
  case BE_KEY_NOT_SET:
    return ("Key object has not been set with key info");
  case BE_KEY_OBJ:
    return ("Invalid key object");
  case BE_KEY_OPERATION_UNKNOWN:
    return ("Unknown operation for a key info type");
  case BE_MEMORY_OBJ:
    return ("Invalid internal memory object");
  case BE_MODULUS_LEN:
    return ("Invalid modulus length in public or private key");
  case BE_NOT_SUPPORTED:
    return ("Unsupported operation requested");
  case BE_OUTPUT_LEN:
    return ("Output data is larger than supplied buffer");
  case BE_OVER_32K:
    return ("Data block exceeds 32,767 bytes");
  case BE_RANDOM_NOT_INITIALIZED:
    return ("Random algorithm has not been initialized");
  case BE_RANDOM_OBJ:
    return ("Invalid algorithm object for the random algorithm");
  case BE_SIGNATURE:
    return ("Invalid signature");
  case BE_WRONG_ALGORITHM_INFO:
    return ("Wrong type of algorithm info");
  case BE_WRONG_KEY_INFO:
    return ("Wrong type of key info");
  default:
    return ((char *)NULL_PTR);
  }
}
