/* Copyright (C) RSA Data Security, Inc. created 1995, 1998.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

/*  This program will generate a series of random values.  You may want
      to rerun this program using the same seed to prove to yourself that
      you get the same random values.  Please refer to Step 4a:  "The
      Random Seed" for a discussion on generating random values.
 */
  
#include <stdio.h>     
#include <string.h>    
#include "aglobal.h"
#include "bsafe.h"

#define NUMBER_OF_RANDOM_BYTES 128

B_ALGORITHM_METHOD *RANDOM_CHOOSER[] = {
  &AM_X931_RANDOM,
  (B_ALGORITHM_METHOD *)NULL_PTR
};

void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));

void main()
{
  B_ALGORITHM_OBJ randomAlgorithm = (B_ALGORITHM_OBJ)NULL_PTR;
  ITEM randomSeed;
  unsigned char *randomByteBuffer = NULL_PTR;
  unsigned int status;
  A_X931_RANDOM_PARAMS x931Params;

  do {
    printf ("Generating random bytes \n");
    printf ("======================= \n");

    /*  Step 1:  Create a random algorithm object */
    if ((status = B_CreateAlgorithmObject (&randomAlgorithm)) != 0)
      break;

    /* Step 2:  Set the random algorithm object to use AI_X931Random.
       Before we can call B_SetAlgorithmInfo, we need to prepare the
       X9.31 parameters.  The A_X931_RANDOM_PARAMS structure
       contains two parameters:  the number of independent streams
       of randomness and an ITEM containing random seed data to be
       divided up among the streams. */

    /* Set the number of streams in the A_X931_RANDOM_PARAMS
       structure. */
    x931Params.numberOfStreams = 6;

    /* In order to obtain a seed, we need to allocate space for it,
       and then request it from the user.  Note that the following
       method of seed gathering is insecure.  A real application would
       use a more secure method of seed gathering to ensure the
       security of the application.  */
    randomSeed.data = T_malloc (384);
    if (randomSeed.data == NULL_PTR)
      break;
    
    do {
      puts ("Enter a random seed (120 bytes minimum, 384 bytes maximum):");
      gets ((char *)randomSeed.data);
      
      randomSeed.len = strlen (randomSeed.data);
    } while (randomSeed.len < 20 * x931Params.numberOfStreams);

    x931Params.seed.data = randomSeed.data;
    x931Params.seed.len = randomSeed.len;

    /* Pass the parameters to the algorithm object in a call to
       B_SetAlgorithmInfo. */
    if ((status = B_SetAlgorithmInfo
         (randomAlgorithm, AI_X931Random, (POINTER)&x931Params)) != 0)
      break;

    /*  Step 3:  Initialize the random algorithm. */
    if ((status = B_RandomInit
         (randomAlgorithm, RANDOM_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Since the random seed has already been passed in via
                 the x931Params structure, we do not have to call
                 B_RandomUpdate(). */

    /*  Step 5:  Generate.  First, prepare a buffer for receiving the
                 random bytes before calling B_GenerateRandomBytes.
     */
    randomByteBuffer = T_malloc (NUMBER_OF_RANDOM_BYTES);
    if ((status = (randomByteBuffer == NULL_PTR)) != 0)
          break;

    T_memset (randomByteBuffer, 0, NUMBER_OF_RANDOM_BYTES);

    if ((status = B_GenerateRandomBytes
         (randomAlgorithm, randomByteBuffer, NUMBER_OF_RANDOM_BYTES,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    printf ("%i bytes of random-generated values: \n",
              NUMBER_OF_RANDOM_BYTES);
    PrintBuf (randomByteBuffer, NUMBER_OF_RANDOM_BYTES);

  } while (0);

  if (status != 0) {
    printf ("Status = %i \n", status);      
    printf ("Generate random bytes failed");
  }

  /*  Step 6:  Remember to destroy all objects, and free all memory
   */
  B_DestroyAlgorithmObject (&randomAlgorithm);
  T_memset (randomSeed.data, 0, randomSeed.len);
  T_free (randomSeed.data);
  T_free (randomByteBuffer);

} /*  end main  */


/*  This procedure will print out what's in the buffer.
 */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */
