/* Copyright (C) RSA Data Security, Inc. created 1998.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

/*  This program demonstrates the use of AI_FeedbackCipher with RC2.
 *  This program can be easily modified to use different feedback modes
 *  or block ciphers, by modifying the fbParams structure below.  For
 *  other related examples, see tdescfb.c, descbc.c, and rc5ofb.c
 */

#include <stdio.h>
#include "aglobal.h"
#include "bsafe.h"

/*  Feedback Mode:  ecb  */
/*  Note that for the given dataToEncrypt, blocks #1 and #4,
 *  and #2 and #5 are identical.
 */
#define FB_MODE "ecb"  

/*  To use other feedback modes, add the new AM's to the chooser and
 *  modify the calls to B_SetAlgorithmInfo.  Specifically, the
 *  feedbackParams field of fbParams.
 */

#define PADDING "pad"  /*  Padding scheme:  pad or nopad  */

/*  Note that for "nopad", the input length must be a multiple of the
 *  block size.  The current dataToEncrypt will work for both padding
 *  options.
 */

#define BLOCK_CIPHER "rc2"
#define KEY_BYTES 24  /*  supply 24 bytes for the RC2 key  */  
#define BLOCK_SIZE 8  /*  8 byte block size  */

/*  To use other block ciphers, add the new AM's to the chooser and
 *  modify the calls to B_SetAlgorithmInfo.  Specifically, modify the
 *  encryptionParams field of fbParams as well as use the appropriate key.
 */

int InitializeRandomAlgorithm PROTO_LIST ((B_ALGORITHM_OBJ *));
void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));

void main()
{
  int status;

  static B_ALGORITHM_METHOD *RC2_ECB_Chooser[] = {
    &AM_ECB_ENCRYPT,
    &AM_ECB_DECRYPT,
    &AM_RC2_ENCRYPT,
    &AM_RC2_DECRYPT,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  unsigned char dataToEncrypt[] = {
    0x45, 0x6e, 0x63, 0x72, 0x79, 0x70, 0x74, 0x20,
    0x74, 0x68, 0x69, 0x73, 0x20, 0x73, 0x65, 0x6e,
    0x74, 0x65, 0x6e, 0x63, 0x65, 0x21, 0x20, 0x20,
    0x45, 0x6e, 0x63, 0x72, 0x79, 0x70, 0x74, 0x20,
    0x74, 0x68, 0x69, 0x73, 0x20, 0x73, 0x65, 0x6e,
    0x74, 0x65, 0x6e, 0x63, 0x65, 0x21, 0x20, 0x00
  };

  unsigned int dataToEncryptLen = sizeof(dataToEncrypt);  

  A_RC2_PARAMS rc2Params;
  ITEM rc2KeyItem;
  unsigned char keyData[KEY_BYTES];
  unsigned int blockLen;
  
  B_KEY_OBJ rc2Key = (B_KEY_OBJ)NULL_PTR;  
  B_ALGORITHM_OBJ randomAlgorithm = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ encryptionObject = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ decryptionObject = (B_ALGORITHM_OBJ)NULL_PTR;

  B_BLK_CIPHER_W_FEEDBACK_PARAMS fbParams;

  unsigned int outputLenUpdate, outputLenFinal, outputLenTotal;
  unsigned int encryptedDataLen, decryptedDataLen;
  unsigned char *encryptedData;
  unsigned char *decryptedData;

  do {
    printf("RC2 Feedback Cipher Example\n");
    printf("===========================\n");
    printf("\nFeedback method:  %s\n\n", FB_MODE);

    if ((status = InitializeRandomAlgorithm(&randomAlgorithm)) != 0)
      break;
    
    /*  Step 1:  Create a key object  */
    if ((status = B_CreateKeyObject(&rc2Key)) != 0)
      break;

    /*  Step 2:  Set the key object  */
    if ((status = B_GenerateRandomBytes(randomAlgorithm, keyData, KEY_BYTES,
                                        (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
    printf("\n%i Random bytes supplied for the key:\n", KEY_BYTES);
    PrintBuf(keyData, KEY_BYTES);

    rc2KeyItem.data = keyData;
    rc2KeyItem.len = KEY_BYTES;

    if ((status = B_SetKeyInfo(rc2Key, KI_Item, (POINTER)&rc2KeyItem)) != 0)
      break;

    printf("Data To Encrypt (%u bytes):\n", dataToEncryptLen);
    printf("%s\nIn hex:\n", dataToEncrypt);
    PrintBuf(dataToEncrypt, dataToEncryptLen);
    
    /*  Step 1:  Create an algorithm object  */
    if ((status = B_CreateAlgorithmObject(&encryptionObject)) != 0)
      break;

    /*  Step 2:  Set the algorithm object with the algorithm information  */
    /*  Set the algorithm information.  Note that this algorithm information
        applies to encryption and decryption, so at this point our algorithm
        object is not yet committed to performing encryption.  */
    rc2Params.effectiveKeyBits = 80;
    blockLen = BLOCK_SIZE;
    
    fbParams.encryptionMethodName = (unsigned char *)BLOCK_CIPHER;
    fbParams.encryptionParams = (POINTER)&rc2Params;
    fbParams.feedbackMethodName = (unsigned char *)FB_MODE;
    fbParams.feedbackParams = (POINTER)&blockLen;
    fbParams.paddingMethodName = (unsigned char *)PADDING;
    fbParams.paddingParams = NULL_PTR;

    if ((status = B_SetAlgorithmInfo(encryptionObject, AI_FeedbackCipher,
                                     (POINTER)&fbParams)) != 0)
      break;

    /*  Step 3:  Initialize the algorithm object to perform encryption  */
    if ((status = B_EncryptInit(encryptionObject, rc2Key, RC2_ECB_Chooser,
                                (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Update  */
    /*  Allocate space to store the encrypted data.  At most, we could end
        up with one block of padding.  However if no padding is used, the
        encrypted data will be the same length as the data to encrypt.  */
    encryptedDataLen = dataToEncryptLen + BLOCK_SIZE;
    encryptedData = T_malloc (encryptedDataLen);
    if ((status = (encryptedData == NULL_PTR)) != 0)
      break;
    
    if ((status = B_EncryptUpdate
	 (encryptionObject, encryptedData, &outputLenUpdate, 
	  encryptedDataLen, dataToEncrypt, dataToEncryptLen,
	  (B_ALGORITHM_OBJ)NULL_PTR, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 5:  Finalize the process  */
    if ((status = B_EncryptFinal
	 (encryptionObject, encryptedData + outputLenUpdate,
	  &outputLenFinal, encryptedDataLen - outputLenUpdate,
	  (B_ALGORITHM_OBJ)NULL_PTR, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    outputLenTotal = outputLenUpdate + outputLenFinal;

    printf("Encrypted Data (%u bytes):\n", outputLenTotal);
    PrintBuf(encryptedData, outputLenTotal);

    printf("Decrypting...\n\n");
    
    /*  Step 1:  Create an algorithm object  */
    if ((status = B_CreateAlgorithmObject(&decryptionObject)) != 0)
      break;

    /*  Step 2:  Set the algorithm object with the algorithm information  */
    /*  We'll use the same algorithm information and initialization vector
        that we used for encryption.  */
    if ((status = B_SetAlgorithmInfo(decryptionObject, AI_FeedbackCipher,
                                     (POINTER)&fbParams)) != 0)
      break;

    /*  Step 3:  Initialize the algorithm object to perform decryption  */
    /*  We'll use the same symmetric key that we used to encrypt with.  */
    if ((status = B_DecryptInit(decryptionObject, rc2Key, RC2_ECB_Chooser,
                                (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Update  */
    /*  Allocate space to store the decrypted data.  The decrypted data can
        not be longer than the encrypted data.  If we are not using padding,
        they will be the same length.  If we are using padding, the decrypted
        data can be anywhere from 1 to 7 bytes shorter than the length of the
        encrypted data.  */
    decryptedDataLen = encryptedDataLen;
    decryptedData = T_malloc (decryptedDataLen);
    if ((status = (decryptedData == NULL_PTR)) != 0)
      break;

    if ((status = B_DecryptUpdate
	 (decryptionObject, decryptedData, &outputLenUpdate, 
	  decryptedDataLen, encryptedData, outputLenTotal,
	  (B_ALGORITHM_OBJ)NULL_PTR, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 5:  Finalize the process  */
    if ((status = B_DecryptFinal
	 (decryptionObject, decryptedData + outputLenUpdate,
	  &outputLenFinal, decryptedDataLen - outputLenUpdate,
	  (B_ALGORITHM_OBJ)NULL_PTR, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    outputLenTotal = outputLenUpdate + outputLenFinal;

    printf("Decrypted Data (%u bytes):\n", outputLenTotal);
    printf("%s\nIn hex:\n", decryptedData);
    PrintBuf(decryptedData, outputLenTotal);

    if ((status = T_memcmp(dataToEncrypt, decryptedData,
                           outputLenTotal)) != 0)
      break;

    printf("The decrypted data matches the original data to encrypt.\n");    
  } while(0);

  if (status != 0)
    printf("RC2 Feedback Cipher Example failed.  Status = %i\n", status);
  
  /*  Step 6:  Destroy  */
  B_DestroyKeyObject(&rc2Key);
  B_DestroyAlgorithmObject(&randomAlgorithm);
  B_DestroyAlgorithmObject(&encryptionObject);
  B_DestroyAlgorithmObject(&decryptionObject);

  if (encryptedData != NULL_PTR) {
    T_memset (encryptedData, 0, encryptedDataLen);
    T_free(encryptedData);
    encryptedData = NULL_PTR;
  }

  if (decryptedData != NULL_PTR) {
    T_memset (decryptedData, 0, decryptedDataLen);
    T_free(decryptedData);
    decryptedData = NULL_PTR;
  }

}

/*  Create a random algorithm
 *  Pass in a pointer to a B_ALGORITHM_OBJ and be sure to call
 *  B_DestroyAlgorithmObject later in the function that calls
 *  this one.
 */
int InitializeRandomAlgorithm (randomAlgorithm)
B_ALGORITHM_OBJ* randomAlgorithm;
{
  int status;

  POINTER randomSeed = NULL_PTR;
  unsigned int randomSeedLen;

  static B_ALGORITHM_METHOD *randomChooser[] = {
    &AM_SHA_RANDOM,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  /*  The following code was pulled from genbytes.c, steps 1-4.  */
  do {
    if ((status = B_CreateAlgorithmObject (randomAlgorithm)) != 0)
      break;

    if ((status = B_SetAlgorithmInfo
                  (*randomAlgorithm, AI_SHA1Random, NULL_PTR)) != 0)
      break;

    if ((status = B_RandomInit
                  (*randomAlgorithm, randomChooser,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    randomSeedLen = 88;
    randomSeed = T_malloc (randomSeedLen);
    if ((status = (randomSeed == NULL_PTR)) != 0)
      break;

    T_memset (randomSeed, 0, randomSeedLen);

    puts ("Enter a random seed");
    if ((status =
	 (NULL_PTR == (unsigned char *)gets ((char *)randomSeed))) != 0)
      break;

    printf("The random seed in hex (%u bytes):\n", randomSeedLen);
    PrintBuf(randomSeed, randomSeedLen);
    
    if ((status = B_RandomUpdate
                  (*randomAlgorithm, randomSeed, randomSeedLen,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
  } while(0);
  
  T_free(randomSeed);

  if (status != 0)
    printf ("Failed to initialize random algorithm.  Status = %i\n", status);
  
  return status;
}    /*  end InitializeRandomAlgorithm   */

/*  This procedure will print out what's in the buffer.
 */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */

