/* Copyright (C) RSA Data Security, Inc. created 1995, 1998.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

/*  This program will generate a DSA keypair, sign with the private
      key, then verify with the public key using AI_DSAWithSHA1.
 */
     
#include <stdio.h>
#include "aglobal.h"
#include "bsafe.h"

B_ALGORITHM_METHOD *DSA_CHOOSER[] = {
  &AM_DSA_PARAM_GEN,
  &AM_DSA_KEY_GEN,
  &AM_DSA_SIGN,
  &AM_DSA_VERIFY,
  &AM_SHA,
  &AM_SHA_RANDOM,
  (B_ALGORITHM_METHOD *)NULL_PTR
};

#define DSA_KEY_BITS   512
#define SIGNATURE_LEN  ((DSA_KEY_BITS + 7) / 8)
#define MAX_SIG_LEN 48

void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));

void main ()
{
  int status;
  
  B_ALGORITHM_OBJ randomAlgorithm = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ dsaParamGenerator = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ dsaKeyGenObj = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ dsaSigner = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ dsaVerifier = (B_ALGORITHM_OBJ)NULL_PTR;
  
  B_KEY_OBJ dsaPublicKey = (B_KEY_OBJ)NULL_PTR;
  B_KEY_OBJ dsaPrivateKey = (B_KEY_OBJ)NULL_PTR;
  
  B_DSA_PARAM_GEN_PARAMS dsaParams;
  
  unsigned char inputData[] = {
    0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,
    0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
    0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,
    0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f
  };
  unsigned int inputDataLen = sizeof(inputData);
  unsigned char signature[MAX_SIG_LEN];
  unsigned int signatureLen;

  do {
    printf ("DSA With SHA1 Digital Signature\n");
    printf ("===============================\n");
    
    printf ("  Generating random bytes \n");
    printf ("  ======================= \n");

    /* Build a random algorithm object.
     */
    if ((status = B_CreateAlgorithmObject (&randomAlgorithm)) != 0)
      break;

    if ((status = B_SetAlgorithmInfo
         (randomAlgorithm, AI_SHA1Random, NULL_PTR)) != 0)
      break;

    if ((status = B_RandomInit
         (randomAlgorithm, DSA_CHOOSER, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    if ((status = B_RandomUpdate
         (randomAlgorithm, inputData, inputDataLen,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    printf ("\n  Generating a DSA Keypair \n");
    printf ("  ======================== \n");

    /* Step 1: Create. */
    if ((status = B_CreateAlgorithmObject (&dsaParamGenerator)) != 0)
      break;
    if ((status = B_CreateAlgorithmObject (&dsaKeyGenObj)) != 0)
      break;
    if ((status = B_CreateKeyObject (&dsaPublicKey)) != 0)
      break;
    if ((status = B_CreateKeyObject (&dsaPrivateKey)) != 0)
      break;

    /* Step 2: Set. */
    dsaParams.primeBits = DSA_KEY_BITS;
    if ((status = B_SetAlgorithmInfo
         (dsaParamGenerator, AI_DSAParamGen, (POINTER)&dsaParams)) != 0)
      break;

    /* Step 3: Init. */
    if ((status = B_GenerateInit
         (dsaParamGenerator, DSA_CHOOSER, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /* Step 5: Generate. */
    if ((status = B_GenerateParameters
         (dsaParamGenerator, dsaKeyGenObj, randomAlgorithm,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /* Step 3: Init. */
    if ((status = B_GenerateInit
         (dsaKeyGenObj, DSA_CHOOSER, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /* Step 5: Generate. */
    if ((status = B_GenerateKeypair
         (dsaKeyGenObj, dsaPublicKey, dsaPrivateKey, randomAlgorithm,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    
    printf ("\n  Signing with DSA \n");
    printf ("  ================ \n");
    printf ("  Data to sign with DSA (%u bytes):\n", inputDataLen);
    PrintBuf (inputData, inputDataLen);
    
    /* Step 1: Create. */
    if ((status = B_CreateAlgorithmObject (&dsaSigner)) != 0)
      break;

    /* Step 2: Set. */
    if ((status = B_SetAlgorithmInfo
         (dsaSigner, AI_DSAWithSHA1, NULL_PTR)) != 0)
      break;

    /* Step 3: Init. */
    if ((status = B_SignInit
         (dsaSigner, dsaPrivateKey, DSA_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /* Step 4: Update. */
    if ((status = B_SignUpdate
         (dsaSigner, inputData, inputDataLen,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /* Step 5: Final. */
    if ((status = B_SignFinal
         (dsaSigner, signature, &signatureLen, MAX_SIG_LEN,
          randomAlgorithm, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    printf ("\n  The DSA digital signature (%u bytes):\n", signatureLen);
    PrintBuf (signature, signatureLen);
    
    printf ("\n  Verifying with DSA \n");
    printf ("  ================== \n");

    /* Step 1: Create. */
    if ((status = B_CreateAlgorithmObject (&dsaVerifier)) != 0)
      break;

    /* Step 2: Set. */
    if ((status = B_SetAlgorithmInfo
         (dsaVerifier, AI_DSAWithSHA1, NULL_PTR)) != 0)
      break;

    /* Step 3: Init. */
    if ((status = B_VerifyInit
         (dsaVerifier, dsaPublicKey, DSA_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /* Step 4: Update. */
    if ((status = B_VerifyUpdate
         (dsaVerifier, inputData, inputDataLen,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /* Step 5: Final. */
    if ((status = B_VerifyFinal
         (dsaVerifier, signature, signatureLen,
          (B_ALGORITHM_OBJ)NULL_PTR, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

  } while (0);
    
  if (status != 0) {
    printf ("Status = %i \n", status);
    printf ("DSA With SHA1 Digital Signature failed");
  }
  else {
    printf ("\nDSA With SHA1 Digital Signature verified.");
  }
  
  /* Step 6: Destroy. */
  B_DestroyAlgorithmObject (&randomAlgorithm);
  B_DestroyAlgorithmObject (&dsaParamGenerator);
  B_DestroyAlgorithmObject (&dsaKeyGenObj);
  B_DestroyAlgorithmObject (&dsaSigner);
  B_DestroyAlgorithmObject (&dsaVerifier);
  B_DestroyKeyObject (&dsaPublicKey);
  B_DestroyKeyObject (&dsaPrivateKey);

}


/*  This procedure will print out what's in the buffer.
 */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */
