/*   Copyright (C) RSA Data Security, Inc. created 1998.
**
**   This file is used to demonstrate how to interface to an
**   RSA Data Security, Inc. licensed development product.
**
**   You have a royalty-free right to use, modify, reproduce and
**   distribute this demonstration file (including any modified
**   version), provided that you agree that RSA Data Security,
**   Inc. has no warranty, implied or otherwise, or liability
**   for this demonstration file or any modified version.
*/

/*  This program will demonstrate creating a digital signature, using
 *  SHA1 to digest and DSA to sign.  Currently, AI_EC_DSAWithDigest
 *  only supports SHA1 for the digest algorithm.
 *
 *  This program will generate an EC DSA keypair, sign data with the
 *  private key, and then verify using the public key.
 *
 *  If you do not wish to use the acceleration table, when compiling
 *  use -DACCEL_TABLE=0, or something similar.
 *
 *  If you want to use the hard-coded EC parameters below, compile with
 *  -DGEN_PARAMS=0, or equivalent.  Otherwise, this program will generate
 *  new EC Parameters using the information in the EC Parameter Generation
 *  Information section.
 */

#include <stdio.h>
#include <time.h>
#include "aglobal.h"
#include "bsafe.h"

#ifndef ACCEL_TABLE
#define ACCEL_TABLE 1
#endif

#ifndef GEN_PARAMS
#define GEN_PARAMS 1
/* Information for EC Parameter Generation  */

/* Currently, the only implementation version supported is 0 */
#define ECPG_VER 0

/* Specify the base field type for the elliptic curve:  FT_FP,
   FT_F2_POLYNOMIAL, or FT_F2_ONB.  See below for restrictions on
   the field element length depending on the field type chosen.  */
#define ECPG_FT FT_FP

/* Specify the length of the prime, basis polynomial, or degree of
   the field -- whichever applies, for fieldElementBits.
   Legal values for field element bits:
     *  with FT_FP or FT_F2_POLYNOMIAL:  64 - 256 bits
     *  with FT_F2_ONB: 65,66,69,74,81,82,83,86,89,90,95,98,99,100,105,
                        106,113,119,130,131,134,135,138,146,148,155,158,
                        162,172,173,174,178,179,180,183,186,189,191,194,
                        196,209,210,221,226,230,231,233,239,243,245,251,254  */
#define ECPG_FE_BITS 160

/* Currently compression is not supported */
#define ECPG_CI CI_NO_COMPRESS

/* Specify the minimum size of the group generated by the base.
   Input of 0 defaults to (fieldElementBits - 7), for minOrderBits */
#define ECPG_MO_BITS 0

/* Specify the maximum size of the second largest prime subgroup of the group.
   Input of 0 defaults to 255, for trialDivBound */
#define ECPG_TDB 0

/* Set if the use of precomputed curves is desired.  This option is only
   available for even polynomial representaions. */
#define ECPG_TL 0
#endif 
/*  end EC Parameter Generation Information  */

int InitializeRandomAlgorithm PROTO_LIST ((B_ALGORITHM_OBJ *));
int InitializeECParamsObj PROTO_LIST ((B_ALGORITHM_OBJ *, B_ALGORITHM_OBJ *));

#if ACCEL_TABLE
int BuildAccelTable PROTO_LIST ((ITEM *, B_ALGORITHM_OBJ *));
int BuildPubKeyAccelTable PROTO_LIST ((ITEM *, B_KEY_OBJ *));
int GenerateECKeys PROTO_LIST ((B_KEY_OBJ *, B_KEY_OBJ *, B_ALGORITHM_OBJ *,
                                B_ALGORITHM_OBJ *, ITEM *));
#else
int GenerateECKeys PROTO_LIST ((B_KEY_OBJ *, B_KEY_OBJ *, B_ALGORITHM_OBJ *,
                                B_ALGORITHM_OBJ *));
#endif

int AllocAndCopyECParamInfo PROTO_LIST ((A_EC_PARAMS *output,
                                         A_EC_PARAMS *input));
int AllocAndCopyECPubKeyInfo PROTO_LIST ((A_EC_PUBLIC_KEY *output,
                                          A_EC_PUBLIC_KEY *input));
void FreeECParamInfo PROTO_LIST ((A_EC_PARAMS *ecParams));
void FreeECPubKeyInfo PROTO_LIST ((A_EC_PUBLIC_KEY *pubKey));
int PrintECParamInfo PROTO_LIST ((B_ALGORITHM_OBJ *));
int PrintECKeyInfo PROTO_LIST ((char *, B_INFO_TYPE,
                                B_KEY_OBJ *, int));
void PrintBufAsCArray PROTO_LIST ((char *, ITEM *));
void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));
int GeneralSurrenderFunction PROTO_LIST ((POINTER handle));

B_ALGORITHM_METHOD *EC_DSA_CHOOSER[] = {
  &AM_SHA,
  &AM_ECFP_DSA_SIGN,
  &AM_ECF2POLY_DSA_SIGN,
  &AM_ECFP_DSA_VERIFY,
  &AM_ECF2POLY_DSA_VERIFY,
  (B_ALGORITHM_METHOD *)NULL_PTR
};

void main()
{
  int status;

  unsigned char dataToSign[] = {
    0x44, 0x6F, 0x6E, 0x27, 0x74, 0x20, 0x72, 0x65,
    0x61, 0x64, 0x20, 0x74, 0x68, 0x69, 0x73, 0x2E,
    0x20, 0x20, 0x49, 0x20, 0x6D, 0x65, 0x61, 0x6E,
    0x20, 0x69, 0x74, 0x2D, 0x2D, 0x73, 0x74, 0x6F,
    0x70, 0x20, 0x72, 0x65, 0x61, 0x64, 0x69, 0x6E,
    0x67, 0x20, 0x74, 0x68, 0x69, 0x73, 0x21, 0x20,
    0x20, 0x54, 0x68, 0x69, 0x73, 0x20, 0x69, 0x73,
    0x20, 0x74, 0x68, 0x65, 0x20, 0x6D, 0x65, 0x73,
    0x73, 0x61, 0x67, 0x65, 0x20, 0x74, 0x68, 0x61,
    0x74, 0x20, 0x77, 0x65, 0x72, 0x65, 0x20, 0x67,
    0x6F, 0x69, 0x6E, 0x67, 0x20, 0x74, 0x6F, 0x20,
    0x73, 0x69, 0x67, 0x6E, 0x20, 0x74, 0x6F, 0x20,
    0x74, 0x65, 0x73, 0x74, 0x20, 0x41, 0x49, 0x5F,
    0x45, 0x43, 0x5F, 0x44, 0x53, 0x41, 0x57, 0x69,
    0x74, 0x68, 0x44, 0x69, 0x67, 0x65, 0x73, 0x74,
    0x2D, 0x2D, 0x62, 0x75, 0x74, 0x20, 0x79, 0x6F,
    0x75, 0x20, 0x73, 0x68, 0x6F, 0x75, 0x6C, 0x64,
    0x6E, 0x27, 0x74, 0x20, 0x72, 0x65, 0x61, 0x64,
    0x20, 0x74, 0x68, 0x69, 0x73, 0x21
  };
  unsigned int dataToSignLen = sizeof(dataToSign);
  
  B_ALGORITHM_OBJ randomAlgorithm = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ ecParamsObj = (B_ALGORITHM_OBJ)NULL_PTR;

  A_EC_PARAMS *ecParamInfo;
  unsigned int orderLen;
  
#if ACCEL_TABLE
  ITEM aTableItem;
  ITEM pubKeyAccelTableItem;    
#endif

  B_KEY_OBJ publicKey = (B_KEY_OBJ)NULL_PTR;
  B_KEY_OBJ privateKey = (B_KEY_OBJ)NULL_PTR;

  B_ALGORITHM_OBJ ecDSASign = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ ecDSAVerify = (B_ALGORITHM_OBJ)NULL_PTR;
  B_DIGEST_SPECIFIER digestInfo;

  unsigned char *signature;
  unsigned int maxSignatureLen;  
  unsigned int signatureLen;  /*  actual length of the signature  */
  
  do {
    printf ("\nEC DSA Signing with SHA1 Example\n");
    printf ("================================\n");
    
    if ((status = InitializeRandomAlgorithm (&randomAlgorithm)) != 0)
      break;
    if ((status = InitializeECParamsObj (&ecParamsObj,
                                         &randomAlgorithm)) != 0)
      break;
    if ((status = PrintECParamInfo (&ecParamsObj)) != 0)
      break;

    /*  Retrieve useful information from the EC Parameters  */
    if ((status = B_GetAlgorithmInfo ((POINTER *)&ecParamInfo, ecParamsObj,
                                      AI_ECParameters)) != 0)
      break;

    orderLen = ecParamInfo->order.len;
    printf("EC Parameters:  Order length = %u bytes\n",
           orderLen);
    
#if ACCEL_TABLE
    if ((status = BuildAccelTable (&aTableItem, &ecParamsObj)) != 0)
      break;
    if ((status = GenerateECKeys (&publicKey, &privateKey, &ecParamsObj,
                                  &randomAlgorithm, &aTableItem)) != 0)
      break;
    if ((status = BuildPubKeyAccelTable (&pubKeyAccelTableItem,
                                         &publicKey)) != 0)
      break;
#else
    if ((status = GenerateECKeys (&publicKey, &privateKey,
                                  &ecParamsObj, &randomAlgorithm)) != 0)
      break;    
#endif
    /*  Print out key information  */
    if ((status = PrintECKeyInfo ("Public Key", KI_ECPublicComponent,
                                  &publicKey, 0)) != 0)
      break;
    if ((status = PrintECKeyInfo ("Private Key", KI_ECPrivateComponent,
                                  &privateKey, 0)) != 0)
      break;

    /*  Step 1:  Create  */
    if ((status = B_CreateAlgorithmObject (&ecDSASign)) != 0)
      break;

    /*  Step 2:  Set  */
    digestInfo.digestInfoType = AI_SHA1;
    digestInfo.digestInfoParams = NULL_PTR;
    if ((status = B_SetAlgorithmInfo (ecDSASign, AI_EC_DSAWithDigest,
                                      (POINTER)&digestInfo)) != 0)
      break;
#if ACCEL_TABLE
    if ((status = B_SetAlgorithmInfo (ecDSASign, AI_ECAcceleratorTable,
                                      (POINTER)&aTableItem)) != 0)
      break;

    T_memset(aTableItem.data, 0, aTableItem.len);    
#endif
    /*  Step 3:  Init  */
    if ((status = B_SignInit (ecDSASign, privateKey, EC_DSA_CHOOSER,
                              (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Update  */
    printf("\nData to sign (%u bytes):\n", dataToSignLen);
    PrintBuf(dataToSign, dataToSignLen);
    if ((status = B_SignUpdate (ecDSASign, dataToSign, dataToSignLen,
                                (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 5:  Final  */
    /*  Allocate space for the signature  */
    maxSignatureLen = (2 * orderLen) + 6;
    printf("Maximum signature length = %u bytes\n", maxSignatureLen);
    signature = T_malloc(maxSignatureLen);
    if ((status = (signature == NULL_PTR)) != 0)
      break;
    
    if ((status = B_SignFinal (ecDSASign, signature, &signatureLen,
                               maxSignatureLen, randomAlgorithm,
                               (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    printf("Signature (%u bytes):\n", signatureLen);
    PrintBuf(signature, signatureLen);

    printf ("Verifying Signature...\n");

    /*  Step 1:  Create  */
    if ((status = B_CreateAlgorithmObject (&ecDSAVerify)) != 0)
      break;

    /*  Step 2:  Set  */
    if ((status = B_SetAlgorithmInfo (ecDSAVerify, AI_EC_DSAWithDigest,
                                      (POINTER)&digestInfo)) != 0)
      break;
#if ACCEL_TABLE
    if ((status = B_SetAlgorithmInfo (ecDSAVerify, AI_ECAcceleratorTable,
                                      (POINTER)&pubKeyAccelTableItem)) != 0)
      break;

    T_memset(pubKeyAccelTableItem.data, 0, pubKeyAccelTableItem.len);    
#endif
    /*  Step 3:  Init  */
    if ((status = B_VerifyInit (ecDSAVerify, publicKey, EC_DSA_CHOOSER,
                                (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Update  */
    if ((status = B_VerifyUpdate (ecDSAVerify, dataToSign, dataToSignLen,
                                  (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 5:  Final  */
    if ((status = B_VerifyFinal (ecDSAVerify, signature, signatureLen,
                                 (B_ALGORITHM_OBJ)NULL_PTR,
                                 (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    printf ("The signature verified.\n");
  } while(0);

  if (status != 0)
    printf("EC DSA example failed.  Status = %i\n", status);

#if ACCEL_TABLE
  T_free(aTableItem.data);
  T_free(pubKeyAccelTableItem.data);
#endif
  T_free(signature);
  B_DestroyAlgorithmObject(&randomAlgorithm);
  B_DestroyAlgorithmObject(&ecParamsObj);
  B_DestroyAlgorithmObject(&ecDSASign);
  B_DestroyAlgorithmObject(&ecDSAVerify);
  B_DestroyKeyObject(&publicKey);
  B_DestroyKeyObject(&privateKey);
}  /*  end main  */

/*  Create a random algorithm
 *  Pass in a pointer to a B_ALGORITHM_OBJ and be sure to call
 *  B_DestroyAlgorithmObject later in the function that calls
 *  this one.
 */
int InitializeRandomAlgorithm (randomAlgorithm)
B_ALGORITHM_OBJ* randomAlgorithm;
{
  int status;

  POINTER randomSeed = NULL_PTR;
  unsigned int randomSeedLen;

  static B_ALGORITHM_METHOD *randomChooser[] = {
    &AM_SHA_RANDOM,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  /*  The following code was pulled from genbytes.c, steps 1-4.  */
  do {
    if ((status = B_CreateAlgorithmObject (randomAlgorithm)) != 0)
      break;

    if ((status = B_SetAlgorithmInfo
                  (*randomAlgorithm, AI_SHA1Random, NULL_PTR)) != 0)
      break;

    if ((status = B_RandomInit
                  (*randomAlgorithm, randomChooser,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    randomSeedLen = 88;
    randomSeed = T_malloc (randomSeedLen);
    if ((status = (randomSeed == NULL_PTR)) != 0)
      break;

    T_memset (randomSeed, 0, randomSeedLen);

    puts ("Enter a random seed");
    if ((status =
	(NULL_PTR == (unsigned char *)gets ((char *)randomSeed))) != 0)
      break;

    printf("The random seed in hex (%u bytes):\n", randomSeedLen);
    PrintBuf(randomSeed, randomSeedLen);

    if ((status = B_RandomUpdate
                  (*randomAlgorithm, randomSeed, randomSeedLen,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
  } while(0);
  
  T_free(randomSeed);

  if (status != 0)
    printf ("Failed to initialize random algorithm.  Status = %i\n", status);
  
  return status;
}    /*  end InitializeRandomAlgorithm   */

/*  Initializes the algorithm object with a set of EC parameters.
 *  Be sure to call B_DestroyAlgorithmObject on ecParamObj after
 *  the do-while in the function that calls this one.
 */
int InitializeECParamsObj (ecParamsObj, randomAlgorithm)
B_ALGORITHM_OBJ *ecParamsObj;
B_ALGORITHM_OBJ *randomAlgorithm;
{
  int status;

#if !GEN_PARAMS
  A_EC_PARAMS stockECParams;

  unsigned char fieldInfo[20] = {
    0xDA, 0xE2, 0x12, 0xCC, 0xEC, 0x6D, 0xFA, 0x73,
    0x17, 0x44, 0x1C, 0xEE, 0x28, 0xF0, 0x42, 0xA3,
    0xDE, 0xD0, 0x4D, 0x7F
  };

  unsigned char coeffA[20] = {
    0xDA, 0xE2, 0x12, 0xCC, 0xEC, 0x6D, 0xFA, 0x73,
    0x17, 0x44, 0x1C, 0xEE, 0x28, 0xF0, 0x42, 0xA3,
    0xDE, 0xD0, 0x4D, 0x7C
  };

  unsigned char coeffB[20] = {
    0xBF, 0x63, 0x40, 0xB3, 0xF8, 0xEF, 0x6A, 0xBC,
    0xD1, 0x9B, 0x56, 0x37, 0x69, 0x85, 0x5B, 0xA0,
    0xA2, 0xAE, 0x84, 0x92
  };

  unsigned char base[41] = {
    0x04, 0x77, 0x79, 0xDC, 0x0B, 0xF7, 0xFA, 0x7E,
    0x52, 0xD1, 0x4C, 0x14, 0x3A, 0x60, 0x7A, 0x46,
    0xE3, 0x6C, 0x7B, 0x7A, 0x7E, 0xD1, 0xA0, 0xC5,
    0x30, 0xA6, 0x2B, 0xF5, 0x4F, 0xA8, 0xE7, 0x6F,
    0x58, 0x64, 0xCC, 0x5A, 0xF3, 0xAB, 0x06, 0x76,
    0x6A
  };

  unsigned char order[20] = {
    0x06, 0x14, 0x80, 0x85, 0xB1, 0x3B, 0xF1, 0x9F,
    0xA4, 0x33, 0xA9, 0x32, 0x42, 0x85, 0x00, 0xFF,
    0x30, 0x43, 0x2E, 0x75
  };

  unsigned char cofactor[1] = {
    0x24
  };

#else
  static B_ALGORITHM_METHOD *paramGenChooser[] = {
    &AM_ECFP_PARAM_GEN,
    &AM_ECF2POLY_PARAM_GEN,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };
  
  B_ALGORITHM_OBJ paramGenObj = (B_ALGORITHM_OBJ)NULL_PTR;
  B_EC_PARAM_GEN_PARAMS paramGenInfo;

  A_SURRENDER_CTX generalSurrenderContext;
  int generalFlag;
#endif
  
  do {
    /*  Step 1:  Create algorithm object to hold EC Parameters */
    if ((status = B_CreateAlgorithmObject(ecParamsObj)) != 0)
      break;

#if !GEN_PARAMS
    /*  Step 2:  Set with old hard-coded EC Parameters */
    stockECParams.version = 0x00;
    stockECParams.fieldType = FT_FP;
    stockECParams.compressIndicator = CI_NO_COMPRESS;
    stockECParams.fieldElementBits = 160;
    stockECParams.fieldInfo.data = fieldInfo;
    stockECParams.fieldInfo.len = sizeof(fieldInfo);
    stockECParams.coeffA.data = coeffA;
    stockECParams.coeffA.len = sizeof(coeffA);
    stockECParams.coeffB.data = coeffB;
    stockECParams.coeffB.len = sizeof(coeffB);
    stockECParams.base.data = base;
    stockECParams.base.len = sizeof(base);
    stockECParams.order.data = order;
    stockECParams.order.len = sizeof(order);
    stockECParams.cofactor.data = cofactor;
    stockECParams.cofactor.len = sizeof(cofactor);
    
    if ((status = B_SetAlgorithmInfo (*ecParamsObj, AI_ECParameters,
                                      (POINTER)&stockECParams)) != 0)
      break;
#else
    /*  Step 1:  Create algorithm object to generate new parameters  */
    if ((status = B_CreateAlgorithmObject(&paramGenObj)) != 0)
      break;

    /*  Step 2:  Set information for generating new EC parameters  */
    paramGenInfo.version = ECPG_VER;
    paramGenInfo.fieldType = ECPG_FT;  
    paramGenInfo.fieldElementBits = ECPG_FE_BITS; 
    paramGenInfo.compressIndicator = ECPG_CI;
    paramGenInfo.minOrderBits = ECPG_MO_BITS;
    paramGenInfo.trialDivBound = ECPG_TDB;
    paramGenInfo.tableLookup = ECPG_TL;
    if ((status = B_SetAlgorithmInfo(paramGenObj, AI_ECParamGen,
                                     (POINTER)&paramGenInfo)) != 0)
      break;

    /*  Step 3:  Init  */
    if ((status = B_GenerateInit(paramGenObj, paramGenChooser,
                                 (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Generate  */
    printf ("\nGenerating new EC Parameters ");
    printf ("(this could take some time)...\n");
    generalSurrenderContext.Surrender = GeneralSurrenderFunction;
    generalSurrenderContext.handle = (POINTER)&generalFlag;
    generalSurrenderContext.reserved = NULL_PTR;
    generalFlag = 0;
    if ((status = B_GenerateParameters(paramGenObj, *ecParamsObj,
                                       *randomAlgorithm,
                                       &generalSurrenderContext)) != 0)
      break;
#endif
  } while(0);

  if (status != 0)
    printf("Failed to generate new EC parameters.  Status = %i\n", status);

#if GEN_PARAMS
  B_DestroyAlgorithmObject (&paramGenObj);
#endif
  
  return status;
}    /*  end InitializeECParamsObj   */

#if ACCEL_TABLE
/*  Build the appropriate acceleration table for the given EC parameters.
 *  accelTableItem.data will be given a pointer to the location of the
 *  allocated space for the acceleration table.  This memory must be freed
 *  after the do-while in the calling function.
 */
int BuildAccelTable (accelTableItem, ecParamsObj)
ITEM *accelTableItem;
B_ALGORITHM_OBJ *ecParamsObj;
{
  int status;

  static B_ALGORITHM_METHOD *ecAccelChooser[] = {
    &AM_ECFP_BLD_ACCEL_TABLE,
    &AM_ECF2POLY_BLD_ACCEL_TABLE,    
    (B_ALGORITHM_METHOD *)NULL_PTR
  };
  
  B_ALGORITHM_OBJ buildTable = (B_ALGORITHM_OBJ)NULL_PTR;
  B_EC_PARAMS paramInfo;
  A_EC_PARAMS *cryptocECParamInfo;
  A_EC_PARAMS ecParamInfo;
  unsigned int maxTableLen;

  A_SURRENDER_CTX generalSurrenderContext;
  int generalFlag;

  do {
    printf ("\nGenerating Acceleration Table (this may take a while)...\n");
 
    /*  Step 1:  Create  */
    if ((status = B_CreateAlgorithmObject(&buildTable)) != 0)
      break;

    /*  Step 2:  Set  */
    /*  Retrieve the EC parameters and place in the algorithm object
        used to build the acceleration table for those parameters  */
    if ((status = B_GetAlgorithmInfo((POINTER *)&cryptocECParamInfo,
                                     *ecParamsObj, AI_ECParameters)) != 0)
      break;

    if ((status = AllocAndCopyECParamInfo(&ecParamInfo,
                                          cryptocECParamInfo)) != 0)
      break;

    paramInfo.parameterInfoType = AI_ECParameters;
    paramInfo.parameterInfoValue = (POINTER)&ecParamInfo;
    
    if ((status = B_SetAlgorithmInfo(buildTable, AI_ECBuildAcceleratorTable,
                                     (POINTER)&paramInfo)) != 0)
      break;

    /*  Step 3:  Init  */
    if ((status = B_BuildTableInit(buildTable, ecAccelChooser,
                                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 5:  Final */
    /*  Allocate sufficient memory to hold the acceleration table  */
    if ((status = B_BuildTableGetBufSize(buildTable, &maxTableLen)) != 0)
      break;
    printf("Max Len of Acceleration Table = %u bytes\n", maxTableLen);
    accelTableItem->data = T_malloc(maxTableLen);
    if ((status = (accelTableItem->data == NULL_PTR)) != 0)
      break;
    
    /*  Build acceleration table  */
    generalSurrenderContext.Surrender = GeneralSurrenderFunction;
    generalSurrenderContext.handle = (POINTER)&generalFlag;
    generalSurrenderContext.reserved = NULL_PTR;
    generalFlag = 0;
    if ((status = B_BuildTableFinal(buildTable, accelTableItem->data,
                                    &(accelTableItem->len), maxTableLen,
                                    &generalSurrenderContext)) != 0)
      break;

    printf("\nAcceleration Table Length = %u bytes\n", accelTableItem->len);
  } while(0);

  if (status != 0)
    printf("Failed to generate acceleration table.  Status = %i\n", status);

  B_DestroyAlgorithmObject(&buildTable);
  FreeECParamInfo(&ecParamInfo);
  
  return status;
}    /*  end BuildAccelTable   */

/*  Build the appropriate acceleration table for the given EC public key.
 *  pubKeyAccelTableItem.data will be given a pointer to the location of the
 *  allocated space for the acceleration table.  This memory must be freed
 *  after the do-while in the calling function.
 */
int BuildPubKeyAccelTable (pubKeyAccelTableItem, publicKey)
ITEM *pubKeyAccelTableItem;
B_KEY_OBJ *publicKey;
{
  int status;

  static B_ALGORITHM_METHOD *ecAccelChooser[] = {
    &AM_ECFP_BLD_PUB_KEY_ACC_TAB,
    &AM_ECF2POLY_BLD_PUB_KEY_ACC_TAB,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };
  
  B_ALGORITHM_OBJ buildTable = (B_ALGORITHM_OBJ)NULL_PTR;
  B_EC_PARAMS paramInfo;
  A_EC_PUBLIC_KEY *cryptocPublicKeyInfo;
  A_EC_PUBLIC_KEY publicKeyInfo;
  unsigned int maxTableLen;

  A_SURRENDER_CTX generalSurrenderContext;
  int generalFlag;
  
  do {
    printf ("\nGenerating Public Key Acceleration Table ");
    printf ("(this may take a while)...\n");
 
    /*  Step 1:  Create  */
    if ((status = B_CreateAlgorithmObject(&buildTable)) != 0)
      break;

    /*  Step 2:  Set  */
    /*  Retrieve the public key information and place in the
        algorithm object used to build the acceleration table
        for that public key  */
    if ((status = B_GetKeyInfo((POINTER *)&cryptocPublicKeyInfo,
                               *publicKey, KI_ECPublic)) != 0)
      break;

    if ((status = AllocAndCopyECPubKeyInfo(&publicKeyInfo,
                                           cryptocPublicKeyInfo)) != 0)
      break;
    
    paramInfo.parameterInfoType = AI_ECPubKey;
    paramInfo.parameterInfoValue = (POINTER)&publicKeyInfo;
    
    if ((status = B_SetAlgorithmInfo(buildTable, AI_ECBuildPubKeyAccelTable,
                                     (POINTER)&paramInfo)) != 0)
      break;

    /*  Step 3:  Init  */
    if ((status = B_BuildTableInit(buildTable, ecAccelChooser,
                                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 5:  Final */
    /*  Allocate sufficient memory to hold the acceleration table  */    
    if ((status = B_BuildTableGetBufSize(buildTable, &maxTableLen)) != 0)
      break;
    printf("Max Len of Public Key Acceleration Table = %u bytes\n",
           maxTableLen);
    pubKeyAccelTableItem->data = T_malloc(maxTableLen);
    if ((status = (pubKeyAccelTableItem->data == NULL_PTR)) != 0)
      break;

    /*  Build acceleration table  */
    generalSurrenderContext.Surrender = GeneralSurrenderFunction;
    generalSurrenderContext.handle = (POINTER)&generalFlag;
    generalSurrenderContext.reserved = NULL_PTR;
    generalFlag = 0;    
    if ((status = B_BuildTableFinal(buildTable, pubKeyAccelTableItem->data,
                                    &(pubKeyAccelTableItem->len), maxTableLen,
                                    &generalSurrenderContext)) != 0)
      break;

    printf("\nPublic Key Acceleration Table = %u bytes\n",
           pubKeyAccelTableItem->len);
  } while(0);

  if (status != 0) {
    printf("Failed to generate public key acceleration table.  ");
    printf("Status = %i\n", status);
  }

  B_DestroyAlgorithmObject(&buildTable);
  FreeECPubKeyInfo(&publicKeyInfo);
  
  return status;
}    /*  end BuildPubKeyAccelTable  */
#endif

/*  Generate a public/private EC keypair.  This procedure needs a pointer
 *  to an initialized random algorithm object.  Be sure to call
 *  B_DestroyKeyObject on the key objects after the do-while in the
 *  function that calls this one.
 */
#if ACCEL_TABLE
int GenerateECKeys (publicKey, privateKey, ecParamsObj,
                    randomAlgorithm, accelTableItem)
B_KEY_OBJ *publicKey;
B_KEY_OBJ *privateKey;
B_ALGORITHM_OBJ *ecParamsObj;
B_ALGORITHM_OBJ *randomAlgorithm;
ITEM *accelTableItem;
#else
int GenerateECKeys (publicKey, privateKey, ecParamsObj, randomAlgorithm)
B_KEY_OBJ *publicKey;
B_KEY_OBJ *privateKey;
B_ALGORITHM_OBJ *ecParamsObj;
B_ALGORITHM_OBJ *randomAlgorithm;
#endif
{
  int status;

  static B_ALGORITHM_METHOD *keyGenChooser[] = {
    &AM_ECFP_KEY_GEN,
    &AM_ECF2POLY_KEY_GEN,    
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  B_ALGORITHM_OBJ ecKeyGen = (B_ALGORITHM_OBJ)NULL_PTR;
  B_EC_PARAMS paramInfo;
  A_EC_PARAMS *cryptocECParamInfo;
  A_EC_PARAMS ecParamInfo;  
  
  do {
    /* Generate an EC Key Pair */
    printf("\nGenerating an EC Key Pair...\n");

    /* Step 1:  Create. */
    if ((status = B_CreateAlgorithmObject (&ecKeyGen)) != 0)
      break;
    if ((status = B_CreateKeyObject (publicKey)) != 0)
      break;
    if ((status = B_CreateKeyObject (privateKey)) != 0)
      break;

    /* Step 2:  Set.  */
    /*  Retrieve the EC parameters and place in the algorithm object
        used to build a key pair for those parameters  */    
    if ((status = B_GetAlgorithmInfo((POINTER *)&cryptocECParamInfo,
                                     *ecParamsObj, AI_ECParameters)) != 0)
      break;

    if ((status = AllocAndCopyECParamInfo(&ecParamInfo,
                                          cryptocECParamInfo)) != 0)
      break;
    
    paramInfo.parameterInfoType = AI_ECParameters;
    paramInfo.parameterInfoValue = (POINTER)&ecParamInfo;
    
    if ((status = B_SetAlgorithmInfo (ecKeyGen, AI_ECKeyGen,
                                      (POINTER)&paramInfo)) != 0)
      break;

#if ACCEL_TABLE
    if ((status = B_SetAlgorithmInfo (ecKeyGen, AI_ECAcceleratorTable,
                                      (POINTER)accelTableItem)) != 0)
      break;
#endif
    
    /* Step 3:  Init.  */
    if ((status = B_GenerateInit (ecKeyGen, keyGenChooser,
                                  (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /* Step 5:  Generate.  */
    if ((status = B_GenerateKeypair
                  (ecKeyGen, *publicKey, *privateKey, *randomAlgorithm,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
  } while(0);

  B_DestroyAlgorithmObject(&ecKeyGen);
  FreeECParamInfo(&ecParamInfo);

  return status;
}    /*  end GenerateECKeys  */

/*  This procedure allocates and copies information from input, a pointer
 *  to an A_EC_PARAMS structure containing memory belonging to CRYPTO-C, to
 *  output, a pointer to an A_EC_PARAMS structure created by the application.
 *  Be sure to call FreeECParamInfo with output when appropriate.
 */
int AllocAndCopyECParamInfo(output, input)
A_EC_PARAMS *output;
A_EC_PARAMS *input;
{
  int status;

  do {
    output->version = input->version;
    
    output->fieldType = input->fieldType;

    output->fieldInfo.len = input->fieldInfo.len;
    output->fieldInfo.data = T_malloc(output->fieldInfo.len);
    if ((status = (output->fieldInfo.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->fieldInfo.data, input->fieldInfo.data,
             output->fieldInfo.len);

    output->coeffA.len = input->coeffA.len;
    output->coeffA.data = T_malloc(output->coeffA.len);
    if ((status = (output->coeffA.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->coeffA.data, input->coeffA.data,
             output->coeffA.len);

    output->coeffB.len = input->coeffB.len;
    output->coeffB.data = T_malloc(output->coeffB.len);
    if ((status = (output->coeffB.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->coeffB.data, input->coeffB.data,
             output->coeffB.len);

    output->base.len = input->base.len;
    output->base.data = T_malloc(output->base.len);
    if ((status = (output->base.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->base.data, input->base.data,
             output->base.len);

    output->order.len = input->order.len;
    output->order.data = T_malloc(output->order.len);
    if ((status = (output->order.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->order.data, input->order.data,
             output->order.len);

    output->cofactor.len = input->cofactor.len;
    output->cofactor.data = T_malloc(output->cofactor.len);
    if ((status = (output->cofactor.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->cofactor.data, input->cofactor.data,
             output->cofactor.len);
    
    output->compressIndicator = input->compressIndicator;
    
    output->fieldElementBits = input->fieldElementBits;
  } while(0);

  if (status != 0)
    printf("AllocAndCopyECParamInfo failed with status %i\n", status);
  
  return status;
}    /*  end AllocAndCopyECParamInfo  */

/*  This procedure allocates and copies information from input, a pointer
 *  to an A_EC_PUBLIC_KEY structure containing memory belonging to CRYPTO-C,
 *  to output, a pointer to an A_EC_PUBLIC_KEY structure created by the
 *  application. Be sure to call FreeECPubKeyInfo with output when appropriate.
 */   
int AllocAndCopyECPubKeyInfo(output, input)
A_EC_PUBLIC_KEY *output;
A_EC_PUBLIC_KEY *input;
{
  int status;

  do {
    output->publicKey.len = input->publicKey.len;
    output->publicKey.data = T_malloc(output->publicKey.len);
    if ((status = (output->publicKey.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->publicKey.data, input->publicKey.data,
             output->publicKey.len);

    if ((status = AllocAndCopyECParamInfo(&(output->curveParams),
                                          &(input->curveParams))) != 0)
      break;
  } while(0);

  if (status != 0)
    printf("AllocAndCopyECPubKeyInfo failed with status %i\n", status);

  return status;
}    /*  end AllocAndCopyECPubKeyInfo  */
  
/*  This procedure takes a pointer to an A_EC_PARAMS structure containing
 *  space allocated by AllocAndCopyECParamInfo and frees all data allocated
 *  with T_malloc.
 */
void FreeECParamInfo(ecParams)
A_EC_PARAMS *ecParams;
{
  T_free(ecParams->fieldInfo.data);
  T_free(ecParams->coeffA.data);
  T_free(ecParams->coeffB.data);
  T_free(ecParams->base.data);
  T_free(ecParams->order.data);
  T_free(ecParams->cofactor.data);
}    /*  end FreeECParamInfo  */

/*  This procedure takes a pointer to an A_EC_PUBLIC_KEY structure containing
 *  space allocated by AllocAndCopyECPubKeyInfo and frees all data allocated
 *  with T_malloc.
 */
void FreeECPubKeyInfo(pubKey)
A_EC_PUBLIC_KEY *pubKey;
{
  T_free(pubKey->publicKey.data);
  FreeECParamInfo(&(pubKey->curveParams));
}    /*  end FreeECPubKeyInfo  */

/*  Given an algorithm object containing EC parameters,
 *  print the information out
 */
int PrintECParamInfo(ecParamsObj)
B_ALGORITHM_OBJ *ecParamsObj;
{
  int status;
  A_EC_PARAMS *cryptocECParamInfo;

  do {
    if ((status = B_GetAlgorithmInfo((POINTER *)&cryptocECParamInfo,
                                     *ecParamsObj, AI_ECParameters)) != 0)
      break;
    printf("\n***EC Parameters Generated:\nversion:  %u\n",
           cryptocECParamInfo->version);
    printf("compress indicator:  ");
    switch (cryptocECParamInfo->compressIndicator) {
      case CI_NO_COMPRESS:
        printf("CI_NO_COMPRESS\n");
        break;
      default:
        printf("Unrecognized compression indicator value:  %i\n",
               cryptocECParamInfo->compressIndicator);
        break;
    }    
    printf("field type:  ");
    switch (cryptocECParamInfo->fieldType) {
      case FT_FP:
        printf("FT_FP\n");
        break;
      case FT_F2_ONB:
        printf("FT_F2_ONB\n");
        break;
      case FT_F2_POLYNOMIAL:
        printf("FT_F2_POLYNOMIAL\n");
        break;
      default:
        printf("Unrecognized field type:  %i\n", cryptocECParamInfo->fieldType);
        break;
    }
    printf("field element length:  %u bits\n",
           cryptocECParamInfo->fieldElementBits);
    PrintBufAsCArray("fieldInfo", &cryptocECParamInfo->fieldInfo);
    PrintBufAsCArray("coeffA", &cryptocECParamInfo->coeffA);
    PrintBufAsCArray("coeffB", &cryptocECParamInfo->coeffB);
    PrintBufAsCArray("base", &cryptocECParamInfo->base);
    PrintBufAsCArray("order", &cryptocECParamInfo->order);
    PrintBufAsCArray("cofactor", &cryptocECParamInfo->cofactor);
  } while(0);

  return status;
}    /*  end PrintECParamInfo  */

/*  This procedure prints out the information in the key object,
 *  depending on the KI provided as the keyInfoType.  If printAsArray
 *  is non-zero, use PrintBufAsCArray to print out the key information.
 *
 *  Valid KIs:
 *  KI_ECPrivateComponent
 *  KI_ECPublicComponent
 */
int PrintECKeyInfo (label, keyInfoType, keyObj, printAsArray)
char *label;
B_INFO_TYPE keyInfoType;
B_KEY_OBJ *keyObj;
int printAsArray;
{
  int status;
  ITEM *cryptocKeyInfo;
  ITEM keyInfo;

  do {
    if ((status = B_GetKeyInfo ((POINTER *)&cryptocKeyInfo, *keyObj,
                                keyInfoType)) != 0)
      break;

    keyInfo.len = cryptocKeyInfo->len;
    keyInfo.data = T_malloc (cryptocKeyInfo->len);
    if ((status = (keyInfo.data == NULL_PTR)) != 0)
      break;
    T_memcpy (keyInfo.data, cryptocKeyInfo->data, keyInfo.len);

    if (printAsArray)
      PrintBufAsCArray(label, &keyInfo);
    else {
      printf("\n%s (%u bytes):\n", label, keyInfo.len);
      PrintBuf(keyInfo.data, keyInfo.len);
    }

    T_memset (keyInfo.data, 0, keyInfo.len);
  } while(0);

  if (status != 0)
    printf("Print EC Key Info failed with status %u\n", status);

  T_free (keyInfo.data);
  
  return status;
}    /*  end PrintECKeyInfo  */

/*  This procedure will print out what's in the buffer as a C array.
 */
void PrintBufAsCArray(label, buffer)
char *label;
ITEM *buffer ;
{
  unsigned int count;

  printf("unsigned char %s[%u] = {\n", label, buffer->len);
  for (count = 0 ; count < buffer->len; count++) {
    if ((count % 8) == 0) {
      printf("  0x%02X", (buffer->data)[count]);
    } else {
      printf(" 0x%02X", (buffer->data)[count]);
    }
    if (count != (buffer->len - 1)) {
      printf(",");
    }
    if (((count % 8) == 7 ) || (count == (buffer->len - 1))) {
      printf("\n");
    }
  }
  printf("};\n");
} /*  end PrintBufAsCArray  */

/*  This procedure will print out what's in the buffer.
 */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */

int GeneralSurrenderFunction (handle)
POINTER handle;
{
  char s[100];
  static time_t currentTime;
  time_t getTime;

  if ((int)*handle == 0) {
    getTime = time(NULL);
    strftime (s, 100, "%H:%M:%S on %A, %d %B %Y", localtime(&getTime));
    printf ("\n%s\n", s);
    printf ("Surrender function ...\n");
    *handle = 1;
    time (&currentTime);
  }
  else {
    time (&getTime);
    if (currentTime != getTime) {
      printf (" .");
      currentTime = getTime;
    }
  }
  return (0);
} /*  end GeneralSurrenderFunction  */
