/* Copyright (C) RSA Data Security, Inc. created 1998.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

/*  This program demonstrates the use of AI_FeedbackCipher with DES.
 *  This program can be easily modified to use different feedback modes
 *  or block ciphers, by modifying the fbParams structure below.  For
 *  other related examples, see descbc.c, rc2ecb.c, and rc5ofb.c
 */

#include <stdio.h>
#include "aglobal.h"
#include "bsafe.h"

/*  Feedback Mode:  cfb or cfb_pipelined  */
#define FB_MODE "cfb"
#define TRANSFER_BITS 8  /*  Transfer size must be 1 or 8 to use "stream"
                             for padding  */

/*  To use other feedback modes, add the new AM's to the chooser and
 *  modify the calls to B_SetAlgorithmInfo.  Specifically, the
 *  feedbackParams field of fbParams.
 */

#define PADDING "pad"  /*  Padding scheme:  pad, nopad, or stream.  */

/*  Note that for "nopad", the input length must be a multiple of the
 *  block size.  The current dataToEncrypt will work for both padding
 *  options.
 */

#define BLOCK_CIPHER "des_ede"
#define KEY_BYTES 24  /*  supply 192 bits for the 3DES key  */  
#define BLOCK_SIZE 8  /*  8 byte block size  */

/*  To use other block ciphers, add the new AM's to the chooser and
 *  modify the calls to B_SetAlgorithmInfo.  Specifically, modify the
 *  encryptionParams field of fbParams as well as use the appropriate key.
 *  Be sure the iv is the same length as the block size.
 */

int InitializeRandomAlgorithm PROTO_LIST ((B_ALGORITHM_OBJ *));
void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));

void main()
{
  int status;

  static B_ALGORITHM_METHOD *TDES_CFB_Chooser[] = {
    &AM_CFB_ENCRYPT,
    &AM_CFB_DECRYPT,
    &AM_CFB_PIPELINED_ENCRYPT,
    &AM_CFB_PIPELINED_DECRYPT,
    &AM_DES_EDE_ENCRYPT,
    &AM_DES_EDE_DECRYPT,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  unsigned char dataToEncrypt[] = {
    0x45, 0x6e, 0x63, 0x72, 0x79, 0x70, 0x74, 0x20,
    0x74, 0x68, 0x69, 0x73, 0x20, 0x73, 0x65, 0x6e,
    0x74, 0x65, 0x6e, 0x63, 0x65, 0x21, 0x20, 0x20,
    0x45, 0x6e, 0x63, 0x72, 0x79, 0x70, 0x74, 0x20,
    0x74, 0x68, 0x69, 0x73, 0x20, 0x73, 0x65, 0x6e,
    0x74, 0x65, 0x6e, 0x63, 0x65, 0x21, 0x20, 0x00
  };

  unsigned int dataToEncryptLen = sizeof(dataToEncrypt);

  ITEM *cryptocDESKey;
  unsigned char keyData[KEY_BYTES];  
  unsigned char ivBytes[BLOCK_SIZE];  
  
  B_KEY_OBJ tdesKey = (B_KEY_OBJ)NULL_PTR;  
  B_ALGORITHM_OBJ randomAlgorithm = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ encryptionObject = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ decryptionObject = (B_ALGORITHM_OBJ)NULL_PTR;

  B_CFB_PARAMS cfbParams;
  B_BLK_CIPHER_W_FEEDBACK_PARAMS fbParams;

  unsigned int outputLenUpdate, outputLenFinal, outputLenTotal;
  unsigned int encryptedDataLen, decryptedDataLen;
  unsigned char *encryptedData = NULL_PTR;
  unsigned char *decryptedData = NULL_PTR;

  do {
    printf("Triple-DES Feedback Cipher Example\n");
    printf("==================================\n");
    printf("\nFeedback method:  %s\n\n", FB_MODE);

    if ((status = InitializeRandomAlgorithm(&randomAlgorithm)) != 0)
      break;
    
    /*  Step 1:  Create a key object  */
    if ((status = B_CreateKeyObject(&tdesKey)) != 0)
      break;

    /*  Step 2:  Set the key object  */
    /*  Generate random bytes for the DES key.  The key is parity
        adjusted when it is copied to the key object.  */
    if ((status = B_GenerateRandomBytes(randomAlgorithm, keyData, KEY_BYTES,
                                        (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
    printf("\nRandom bytes supplied for the key:\n");
    PrintBuf(keyData, KEY_BYTES);

    if ((status = B_SetKeyInfo(tdesKey, KI_DES24Strong,
                               (POINTER)keyData)) != 0)
      break;

    /*  tdesKey now contains the parity-adjusted DES key.  This symmetric key
        will be used for encryption and decryption.  */

    /*  Retrieve the parity-adjusted key, just for kicks.  The pointer we
        get back points to memory that belongs to CRYPTO-C.  So in a real
        application, if we wanted to transport the key for example, we'd
        have to make our own copy of the information.  */
    if ((status = B_GetKeyInfo((POINTER *)&cryptocDESKey, tdesKey,
                               KI_Item)) != 0)
      break;
    printf("Parity-adjusted triple-DES key:\n");
    PrintBuf(cryptocDESKey->data, cryptocDESKey->len);

    printf("Data To Encrypt (%u bytes):\n", dataToEncryptLen);
    printf("%s\nIn hex:\n", dataToEncrypt);
    PrintBuf(dataToEncrypt, dataToEncryptLen);
    
    /*  Step 1:  Create an algorithm object  */
    if ((status = B_CreateAlgorithmObject(&encryptionObject)) != 0)
      break;

    /*  Step 2:  Set the algorithm object with the algorithm information  */
    /*  Set the algorithm information.  Note that this algorithm information
        applies to encryption and decryption, so at this point our algorithm
        object is not yet committed to performing encryption.  */
    if ((status = B_GenerateRandomBytes(randomAlgorithm, ivBytes, BLOCK_SIZE,
                                        (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
    
    cfbParams.ivItem.data = ivBytes;
    cfbParams.ivItem.len = BLOCK_SIZE;
    cfbParams.transferSize = TRANSFER_BITS;

    printf("Initialization vector (%u bytes):\n", cfbParams.ivItem.len);
    PrintBuf(cfbParams.ivItem.data, cfbParams.ivItem.len);        
    
    fbParams.encryptionMethodName = (unsigned char *)BLOCK_CIPHER;
    fbParams.encryptionParams = NULL_PTR;
    fbParams.feedbackMethodName = (unsigned char *)FB_MODE;
    fbParams.feedbackParams = (POINTER)&cfbParams;
    fbParams.paddingMethodName = (unsigned char *)PADDING;
    fbParams.paddingParams = NULL_PTR;

    if ((status = B_SetAlgorithmInfo(encryptionObject, AI_FeedbackCipher,
                                     (POINTER)&fbParams)) != 0)
      break;

    /*  Step 3:  Initialize the algorithm object to perform encryption  */
    if ((status = B_EncryptInit(encryptionObject, tdesKey, TDES_CFB_Chooser,
                                (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Update  */
    /*  Allocate space to store the encrypted data.  At most, we could end
        up with one block of padding.  However if no padding is used, the
        encrypted data will be the same length as the data to encrypt.  */
    encryptedDataLen = dataToEncryptLen + BLOCK_SIZE;
    encryptedData = T_malloc (encryptedDataLen);
    if ((status = (encryptedData == NULL_PTR)) != 0)
      break;
    
    if ((status = B_EncryptUpdate
	 (encryptionObject, encryptedData, &outputLenUpdate, 
	  encryptedDataLen, dataToEncrypt, dataToEncryptLen,
	  (B_ALGORITHM_OBJ)NULL_PTR, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 5:  Finalize the process  */
    if ((status = B_EncryptFinal
	 (encryptionObject, encryptedData + outputLenUpdate,
	  &outputLenFinal, encryptedDataLen - outputLenUpdate,
	  (B_ALGORITHM_OBJ)NULL_PTR, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    outputLenTotal = outputLenUpdate + outputLenFinal;

    printf("Encrypted Data (%u bytes):\n", outputLenTotal);
    PrintBuf(encryptedData, outputLenTotal);

    printf("Decrypting...\n\n");
    
    /*  Step 1:  Create an algorithm object  */
    if ((status = B_CreateAlgorithmObject(&decryptionObject)) != 0)
      break;

    /*  Step 2:  Set the algorithm object with the algorithm information  */
    /*  We'll use the same algorithm information and initialization vector
        that we used for encryption.  */
    if ((status = B_SetAlgorithmInfo(decryptionObject, AI_FeedbackCipher,
                                     (POINTER)&fbParams)) != 0)
      break;

    /*  Step 3:  Initialize the algorithm object to perform decryption  */
    /*  We'll use the same symmetric key that we used to encrypt with.  */
    if ((status = B_DecryptInit(decryptionObject, tdesKey, TDES_CFB_Chooser,
                                (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Update  */
    /*  Allocate space to store the decrypted data.  The decrypted data can
        not be longer than the encrypted data.  If we are not using padding,
        they will be the same length.  If we are using padding, the decrypted
        data can be anywhere from 1 to 7 bytes shorter than the length of the
        encrypted data.  */
    decryptedDataLen = encryptedDataLen;
    decryptedData = T_malloc (decryptedDataLen);
    if ((status = (decryptedData == NULL_PTR)) != 0)
      break;

    if ((status = B_DecryptUpdate
	 (decryptionObject, decryptedData, &outputLenUpdate,
	  decryptedDataLen, encryptedData, outputLenTotal,
	  (B_ALGORITHM_OBJ)NULL_PTR, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 5:  Finalize the process  */
    if ((status = B_DecryptFinal
	 (decryptionObject, decryptedData + outputLenUpdate,
	  &outputLenFinal, decryptedDataLen - outputLenUpdate,
	  (B_ALGORITHM_OBJ)NULL_PTR, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    outputLenTotal = outputLenUpdate + outputLenFinal;
    
    printf("Decrypted Data (%u bytes):\n", outputLenTotal);
    printf("%s\nIn hex:\n", decryptedData);
    PrintBuf(decryptedData, outputLenTotal);

    if ((status = T_memcmp(dataToEncrypt, decryptedData,
                           outputLenTotal)) != 0)
      break;

    printf("The decrypted data matches the original data to encrypt.\n");
    
  } while(0);

  if (status != 0)
    printf("Triple-DES Feedback Cipher Example failed.  Status = %i\n",
           status);
  
  /*  Step 6:  Destroy  */
  B_DestroyKeyObject(&tdesKey);
  B_DestroyAlgorithmObject(&randomAlgorithm);
  B_DestroyAlgorithmObject(&encryptionObject);
  B_DestroyAlgorithmObject(&decryptionObject);

  if (encryptedData != NULL_PTR) {
    T_memset (encryptedData, 0, encryptedDataLen);
    T_free (encryptedData);
    encryptedData = NULL_PTR;
  }

  if (decryptedData != NULL_PTR) {
    T_memset (decryptedData, 0, decryptedDataLen);
    T_free(decryptedData);
    decryptedData = NULL_PTR;
  }
}

/*  Create a random algorithm
 *  Pass in a pointer to a B_ALGORITHM_OBJ and be sure to call
 *  B_DestroyAlgorithmObject later in the function that calls
 *  this one.
 */
int InitializeRandomAlgorithm (randomAlgorithm)
B_ALGORITHM_OBJ* randomAlgorithm;
{
  int status;

  POINTER randomSeed = NULL_PTR;
  unsigned int randomSeedLen;

  static B_ALGORITHM_METHOD *randomChooser[] = {
    &AM_SHA_RANDOM,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  /*  The following code was pulled from genbytes.c, steps 1-4.  */
  do {
    if ((status = B_CreateAlgorithmObject (randomAlgorithm)) != 0)
      break;

    if ((status = B_SetAlgorithmInfo
                  (*randomAlgorithm, AI_SHA1Random, NULL_PTR)) != 0)
      break;

    if ((status = B_RandomInit
                  (*randomAlgorithm, randomChooser,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    randomSeedLen = 88;
    randomSeed = T_malloc (randomSeedLen);
    if ((status = (randomSeed == NULL_PTR)) != 0)
      break;

    T_memset (randomSeed, 0, randomSeedLen);

    puts ("Enter a random seed");
    if ((status =
	 (NULL_PTR == (unsigned char *)gets ((char *)randomSeed))) != 0)
      break;

    printf("The random seed in hex (%u bytes):\n", randomSeedLen);
    PrintBuf(randomSeed, randomSeedLen);    

    if ((status = B_RandomUpdate
                  (*randomAlgorithm, randomSeed, randomSeedLen,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
  } while(0);
  
  T_free(randomSeed);

  if (status != 0)
    printf ("Failed to initialize random algorithm.  Status = %i\n", status);
  
  return status;
}    /*  end InitializeRandomAlgorithm   */

/*  This procedure will print out what's in the buffer.
 */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */
