/* 
   Copyright (C) RSA Data Security, Inc. created 1995, 1998.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

/* This program will generate a RSA keypair, encrypt with the public key,
 * then decrypt with the private key -- i.e. Digital Envelope using
 * AI_PKCS_OAEP_RSAPublic and AI_PKCS_OAEP_RSAPrivate to encrypt and 
 * decrypt randomly generated data using OAEP as defined in PKCS #1 v2.  
 *
 * The Optimal Asymmetric Encryption Padding (OAEP) scheme is used as a
 * superior padding scheme to thrwart certain attacks against the protocols
 * using Digital Envelopes.  For more information, see
 *
 * http://www.rsa.com/rsalabs/pubs/PKCS/ 
 *
 * For more information on the issues surrounding PKCS #1 v2.
 *
 * 19NOV98 JAG v4.1.1
 */

#include <stdio.h>   
#include <string.h>  
#include <time.h>
#include "aglobal.h"
#include "bsafe.h"
#include "demochos.h"

#define RSA_MODULUS_BITS 1024

/* PKCS #1 OAEP has 42 bytes of padding */
#define BLOCK_OVERHEAD 42

/* Helper function prototypes. */
int InitializeRandomAlgorithm PROTO_LIST
     ((B_ALGORITHM_OBJ*, POINTER, unsigned int));
int GenerateRSAKeypair PROTO_LIST 
     ((B_ALGORITHM_OBJ *, B_KEY_OBJ *, B_KEY_OBJ *, int));
void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));

void main()
{

  /* Random algorithm objects. */
  B_ALGORITHM_OBJ randomAlgorithm = (B_ALGORITHM_OBJ)NULL_PTR;

  /* Public and private key objects. */
  B_KEY_OBJ publicKey = (B_KEY_OBJ)NULL_PTR;
  B_KEY_OBJ privateKey = (B_KEY_OBJ)NULL_PTR;

  /* Encryption and decryption objects. */
  B_ALGORITHM_OBJ rsaEncryptor = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ rsaDecryptor = (B_ALGORITHM_OBJ)NULL_PTR;

  unsigned char *dataToEncrypt;
  unsigned int dataToEncryptLen;
  unsigned char *encryptedData;
  unsigned char *decryptedData;
  unsigned char *updateBuffer;
  unsigned int encryptedDataLen, decryptedDataLen;
  unsigned int updateBufferLen;
  unsigned int status, blockSize, blockPayload;

  int randomSeedLen;
  POINTER randomSeed = NULL_PTR;

  do {
      /* Gather random seed from user input and then initialize random 
       * algorithms.
       */
      randomSeedLen = 256;
      randomSeed = T_malloc (randomSeedLen);
      if ((status = (randomSeed == NULL_PTR)) != 0)
         break;

      T_memset (randomSeed, 0, randomSeedLen);

      puts ("Enter a random seed");
      if ((status =
          (NULL_PTR == (unsigned char *)gets ((char *)randomSeed))) != 0)
      break;

	  if ((status =InitializeRandomAlgorithm 
		  (&randomAlgorithm,randomSeed,randomSeedLen)) != 0) 
      break;
	  
      /* Determine our encryption block size, and the payload (data) that we
       * can encrypt in each block.
       */
      blockSize =  ( RSA_MODULUS_BITS + 7) / 8;
      blockPayload = blockSize - BLOCK_OVERHEAD;

      /* Encrypted data is allocated to the max message size.
       */
      dataToEncrypt = T_malloc(blockPayload);
      dataToEncryptLen = blockPayload;
	  	  
      encryptedData = T_malloc(blockSize);
      updateBuffer = T_malloc(blockSize);

      /* Decrypted data is allocated to the padded max message size. 
       */
      decryptedData = T_malloc(blockPayload);
	
      /* For this example, we'll envelope some randomly generated bytes. 
       */
      if ((status = B_GenerateRandomBytes
          (randomAlgorithm, dataToEncrypt, dataToEncryptLen, 
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

      printf ("Data to Encrypt and Pad: \n");
      printf ("============================= \n");
	  
	  PrintBuf(dataToEncrypt, dataToEncryptLen);

      /* Generate a RSA Keypair.  For more information on generating RSA
       * Key pairs, see rsasign.c.
       */
      if ((status = GenerateRSAKeypair (&randomAlgorithm, &publicKey,
                                      &privateKey, RSA_MODULUS_BITS)) != 0)
      break;

      printf ("Performing PKCS V2 Encryption \n");
      printf ("============================= \n");

      /*  Step 1:  Creating an Algorithm Object */
      if ((status = B_CreateAlgorithmObject (&rsaEncryptor)) != 0)
      break;

      /*  Step 2:  Set the algorithm object to AI_PKCS_OAEP_RSAPublic */
      if ((status = B_SetAlgorithmInfo
         (rsaEncryptor, AI_PKCS_OAEP_RSAPublic, NULL_PTR)) != 0)
      break;

      /*  Step 3:  Init -- encrypt with the recipient's public key */
      if ((status = B_EncryptInit
         (rsaEncryptor, publicKey,DEMO_ALGORITHM_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

      /*  Step 4:  Update */
      if ((status = B_EncryptUpdate
          (rsaEncryptor, encryptedData, &encryptedDataLen, blockSize,
          (unsigned char *)dataToEncrypt, dataToEncryptLen,
           randomAlgorithm, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

      /*  Step 5:  Final */
      if ((status = B_EncryptFinal
         (rsaEncryptor, encryptedData,
          &encryptedDataLen, blockSize,
          randomAlgorithm, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

      printf ("Encrypted Data with Padding: \n");
      printf ("============================= \n");
	  
	  PrintBuf(encryptedData, encryptedDataLen);


      printf ("\nPerforming PKCS V2 Decryption \n");
      printf ("================================ \n");

      /*  Step 1:  Creating an Algorithm Object */
      if ((status = B_CreateAlgorithmObject (&rsaDecryptor)) != 0)
      break;

      /*  Step 2:  Set the algorithm object to AI_PKCS_OAEP_RSAPublic */
      if ((status = B_SetAlgorithmInfo
         (rsaDecryptor, AI_PKCS_OAEP_RSAPrivate, NULL_PTR)) != 0)
      break;

      /*  Step 3:  Init  */
      if ((status = B_DecryptInit
         (rsaDecryptor, privateKey, DEMO_ALGORITHM_CHOOSER,
         (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

      /*  Step 4:  Update */
      updateBufferLen = 0;
      if ((status = B_DecryptUpdate
         (rsaDecryptor, updateBuffer, &updateBufferLen, blockPayload,
         (unsigned char *)encryptedData, encryptedDataLen,
          randomAlgorithm, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

      /*  Step 5:  Final */
      if ((status = B_DecryptFinal
         (rsaDecryptor, decryptedData, &decryptedDataLen, blockPayload,
          randomAlgorithm, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

      printf ("Decrypted Data: \n");
      printf ("============================= \n");
	  
	  PrintBuf(encryptedData, encryptedDataLen);

      /* The decrypted data should match the plaintext data that was 
       * originally generated.  If they match, the envelope verified
       * correctly.
       */
      if (T_memcmp (dataToEncrypt, decryptedData, decryptedDataLen)){
		printf ("Envelope did not verify.\n");
		break;
      } else printf("Envelope verified.\n");


  } while (0);
   
  if (status) {
	printf("\nExit with bad status = %d ", status); 
  }


  /* Free the memory used by these objects.  The T_ functions are
   * defined in tstdlib.c included in the samples/source/general
   * directory with CRYPTO-C.  We'll also zero out the memory for
   * security precautions.
   */
    
  if (dataToEncrypt != NULL_PTR) {
       T_memset(dataToEncrypt, 0, dataToEncryptLen);
       T_free(dataToEncrypt);
  }

  if (encryptedData != NULL_PTR) {
       T_memset(encryptedData, 0, encryptedDataLen);
       T_free(encryptedData);
  }

  if (updateBuffer != NULL_PTR) {
       T_memset(updateBuffer, 0, updateBufferLen);
       T_free(updateBuffer);
  }

  if (decryptedData != NULL_PTR) {
       T_memset(decryptedData, 0, decryptedDataLen);
       T_free(decryptedData);
  }


  /*  
  Step 6:  Destroy
  */
  B_DestroyAlgorithmObject (&randomAlgorithm);
  B_DestroyKeyObject (&privateKey);
  B_DestroyKeyObject (&publicKey);
  B_DestroyAlgorithmObject (&rsaDecryptor);
  B_DestroyAlgorithmObject (&rsaEncryptor);

} /*  end main  */


int InitializeRandomAlgorithm (randomAlgorithm,randomSeed,randomSeedLen)
B_ALGORITHM_OBJ* randomAlgorithm;
POINTER randomSeed;
unsigned int randomSeedLen;
{
  int status;

  static B_ALGORITHM_METHOD *randomChooser[] = {
    &AM_MD5_RANDOM,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  /*  The following code was pulled from genbytes.c, steps 1-4.  */
  do {
    if ((status = B_CreateAlgorithmObject (randomAlgorithm)) != 0)
      break;

    if ((status = B_SetAlgorithmInfo
                  (*randomAlgorithm, AI_MD5Random, NULL_PTR)) != 0)
      break;

    if ((status = B_RandomInit
                  (*randomAlgorithm, randomChooser,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    if ((status = B_RandomUpdate
                  (*randomAlgorithm, randomSeed, randomSeedLen,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
  } while(0);
  
  if (status != 0)
    printf ("Failed to initialize random algorithm.  Status = %i\n", status);
  
  return status;
}    /*  end InitializeRandomAlgorithm   */


/*
    Create an RSA public/private keypair
    randomAlgorithm is a pointer to an already-initialized random algorithm
    object.  Be sure to call B_DestroyKeyObject on the public and private
    keys.
 */
int GenerateRSAKeypair (randomAlgorithm, publicKey, privateKey, modulusBits)
B_ALGORITHM_OBJ *randomAlgorithm;
B_KEY_OBJ *publicKey;
B_KEY_OBJ *privateKey;
int modulusBits;
{
  static B_ALGORITHM_METHOD *RSA_KEYGEN_CHOOSER[] = {
    &AM_RSA_KEY_GEN,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  B_ALGORITHM_OBJ keypairGenerator = (B_ALGORITHM_OBJ)NULL_PTR;
  A_RSA_KEY_GEN_PARAMS keygenParams;

  static unsigned char f4Data[] = {0x01, 0x00, 0x01};

  int status;

  do {
    printf ("Generating a Keypair \n");
    printf ("==================== \n");

    /*  Step 1:  Create algorithm object */
    if ((status = B_CreateAlgorithmObject (&keypairGenerator)) != 0)
      break;

    /*  Step 2:  Set algorithm object to AI_RSAKeyGen */
    keygenParams.modulusBits = modulusBits;
    keygenParams.publicExponent.data = f4Data;
    keygenParams.publicExponent.len = sizeof(f4Data);
    if ((status = B_SetAlgorithmInfo (keypairGenerator, AI_RSAKeyGen,
                                      (POINTER)&keygenParams)) != 0)
      break;

    /*  Step 3:  Init */
    if ((status = B_GenerateInit
                  (keypairGenerator, RSA_KEYGEN_CHOOSER,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  no step 4 (Update) in generating a keypair */

    /*  Step 5:  Generate
     */
    if ((status = B_CreateKeyObject (publicKey)) != 0)
      break;
    if ((status = B_CreateKeyObject (privateKey)) != 0)
      break;

    if ((status = B_GenerateKeypair
                  (keypairGenerator, *publicKey, *privateKey,
                   *randomAlgorithm, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

  } while(0);

  if (status != 0)
    printf ("Failed to generate RSA keypair.  Status = %i\n", status);

  B_DestroyAlgorithmObject (&keypairGenerator);

  return status;
}     /*  end GenerateRSAKeypair  */

/*  This procedure will print out what's in the buffer. */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */
