/* Copyright (C) RSA Data Security, Inc. created 1995, 1998.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

/*  This program will generate a RSA keypair, encrypt with the public key,
      then decrypt with the private key -- i.e. Digital Envelope.
 */

#include <stdio.h>   
#include <string.h>  
#include <time.h>
#include "aglobal.h"
#include "bsafe.h"

B_ALGORITHM_METHOD *RSA_SAMPLE_CHOOSER[] = {
  &AM_SHA_RANDOM,
  &AM_RSA_KEY_GEN,
  &AM_RSA_ENCRYPT,
  &AM_RSA_DECRYPT,
  &AM_RSA_CRT_ENCRYPT,
  &AM_RSA_CRT_DECRYPT,
  (B_ALGORITHM_METHOD *)NULL_PTR
};

#define RSA_MODULUS_BITS 512
#define BLOCK_SIZE ((RSA_MODULUS_BITS + 7) / 8)

int GeneralSurrenderFunction PROTO_LIST ((POINTER handle));

void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));

void main()
{
  B_ALGORITHM_OBJ randomAlgorithm = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ keypairGenerator = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ rsaEncryptor = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ rsaDecryptor = (B_ALGORITHM_OBJ)NULL_PTR;
  B_KEY_OBJ publicKey = (B_KEY_OBJ)NULL_PTR;
  B_KEY_OBJ privateKey = (B_KEY_OBJ)NULL_PTR;

  POINTER randomSeed = NULL_PTR;
  unsigned int randomSeedLen;
  
  A_RSA_KEY_GEN_PARAMS keygenParams;
  
  static unsigned char f4Data[3] = {0x01, 0x00, 0x01};

  A_SURRENDER_CTX generalSurrenderContext;
  int generalFlag;

  ITEM *cryptocPublicKeyBER;
  ITEM myPublicKeyBER;

  ITEM *getPrivateKey;
  ITEM privateKeyBER;
  
  static unsigned char dataToEncryptWithRSA[] =
    {0x4a, 0x72, 0x55, 0x36, 0xda, 0x2f, 0xb9, 0x51};

  unsigned int dataToEncryptLen = sizeof(dataToEncryptWithRSA);

  unsigned char encryptedData[BLOCK_SIZE];
  unsigned char decryptedData[BLOCK_SIZE];
  unsigned int outputLenUpdate, outputLenFinal, outputLenTotal;

  unsigned int status;

  myPublicKeyBER.data = NULL_PTR;
  privateKeyBER.data = NULL_PTR;

  generalSurrenderContext.Surrender = GeneralSurrenderFunction;
  generalSurrenderContext.handle = (POINTER)&generalFlag;
  generalSurrenderContext.reserved = NULL_PTR;

  do {
    printf ("\nRSA PKCS Algorithm \n");
    printf ("================== \n");
    printf ("Maximum allowable size for input data = %i\n", BLOCK_SIZE - 11);
    printf ("Input data (%u bytes):\n", dataToEncryptLen);
    PrintBuf (dataToEncryptWithRSA, dataToEncryptLen);

    /*  ========================================================  */
    /*  The following code was pulled from genbytes.c, steps 1-4.
     */
    printf ("Generating random bytes \n");
    printf ("======================= \n");

    if ((status = B_CreateAlgorithmObject (&randomAlgorithm)) != 0)
      break;

    if ((status = B_SetAlgorithmInfo
         (randomAlgorithm, AI_SHA1Random, NULL_PTR)) != 0)
      break;

    if ((status = B_RandomInit
         (randomAlgorithm, RSA_SAMPLE_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    randomSeedLen = 256;
    randomSeed = T_malloc (randomSeedLen);
    if ((status = (randomSeed == NULL_PTR)) != 0)
         break;

    T_memset (randomSeed, 0, randomSeedLen);

    puts ("Enter a random seed");
    if ((status =
	 (NULL_PTR == (unsigned char *)gets ((char *)randomSeed))) != 0)
      break;

    if ((status = B_RandomUpdate
         (randomAlgorithm, randomSeed, randomSeedLen,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    printf ("   Generating a Keypair \n");
    printf ("   ==================== \n");

    /*  Step 1:  Create algorithm object */
    if ((status = B_CreateAlgorithmObject (&keypairGenerator)) != 0)
      break;

    /*  Step 2:  Set algorithm object to AI_RSAKeyGen
     */
    keygenParams.modulusBits = RSA_MODULUS_BITS;
    keygenParams.publicExponent.data = f4Data;
    keygenParams.publicExponent.len = 3;

    if ((status = B_SetAlgorithmInfo
         (keypairGenerator, AI_RSAKeyGen, (POINTER)&keygenParams)) != 0)
      break;

    /*  Step 3:  Init */
    if ((status = B_GenerateInit
         (keypairGenerator, RSA_SAMPLE_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  no step 4 (Update) in generating a keypair */

    /*  Step 5:  Generate
     */
    if ((status = B_CreateKeyObject (&publicKey)) != 0)
      break;
    if ((status = B_CreateKeyObject (&privateKey)) != 0)
      break;

    /*  generalFlag is for the surrender function */
    generalFlag = 0;
    if ((status = B_GenerateKeypair
         (keypairGenerator, publicKey, privateKey,
          randomAlgorithm, &generalSurrenderContext)) != 0)
      break;
    
    printf ("\n   Distributing Your Public Key \n");
    printf ("   ============================ \n");

    myPublicKeyBER.data = NULL_PTR;
    if ((status = B_GetKeyInfo
         ((POINTER *)&cryptocPublicKeyBER, publicKey, KI_RSAPublicBER)) != 0)
      break;

    myPublicKeyBER.len = cryptocPublicKeyBER->len;
    myPublicKeyBER.data = T_malloc (myPublicKeyBER.len);

    if ((status = (myPublicKeyBER.data == NULL_PTR)) != 0)
      break;
    T_memcpy (myPublicKeyBER.data, cryptocPublicKeyBER->data,
              myPublicKeyBER.len);

    printf ("The public key DER-encoded as X.509 SubjectPublicKeyInfo type");
    printf (" (%u bytes):\n", myPublicKeyBER.len);
    PrintBuf (myPublicKeyBER.data, myPublicKeyBER.len);

    printf ("   Encrypting with the RSA Public Key \n");
    printf ("   ================================== \n");

    /*  Step 1:  Creating an Algorithm Object */
    if ((status = B_CreateAlgorithmObject (&rsaEncryptor)) != 0)
      break;

    /*  Step 2:  Set the algorithm object to AI_PKCS_RSAPublic */
    if ((status = B_SetAlgorithmInfo
         (rsaEncryptor, AI_PKCS_RSAPublic, NULL_PTR)) != 0)
      break;

    /*  Step 3:  Init -- encrypt with the recipient's public key */
    if ((status = B_EncryptInit
         (rsaEncryptor, publicKey, RSA_SAMPLE_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Update */
    generalFlag = 0;
    if ((status = B_EncryptUpdate
                  (rsaEncryptor, encryptedData, &outputLenUpdate, BLOCK_SIZE,
                   (unsigned char *)dataToEncryptWithRSA, dataToEncryptLen,
                   randomAlgorithm, &generalSurrenderContext)) != 0)
      break;

    /*  Step 5:  Final
     */
    generalFlag = 0;
    if ((status = B_EncryptFinal
         (rsaEncryptor, encryptedData + outputLenUpdate,
          &outputLenFinal, BLOCK_SIZE - outputLenUpdate,
          randomAlgorithm, &generalSurrenderContext)) != 0)
      break;
    
    outputLenTotal = outputLenUpdate + outputLenFinal;
    printf ("\nThe encrypted data (%u bytes):\n", outputLenTotal);
    PrintBuf (encryptedData, outputLenTotal);

    printf ("   Decrypting with the RSA Private Key \n");
    printf ("   =================================== \n");

    privateKeyBER.data = NULL_PTR;
    if ((status = B_GetKeyInfo
         ((POINTER *)&getPrivateKey, privateKey,
          KI_PKCS_RSAPrivateBER)) != 0)
      break;

    privateKeyBER.len = getPrivateKey->len;
    privateKeyBER.data = T_malloc (privateKeyBER.len);

    if ((status = (privateKeyBER.data == NULL_PTR)) != 0)
      break;
    T_memcpy (privateKeyBER.data, getPrivateKey->data,
              privateKeyBER.len);

    printf ("The private key encoded as a PKCS #8 PrivateKeyInfo type");
    printf (" (%u bytes):\n", privateKeyBER.len);
    PrintBuf (privateKeyBER.data, privateKeyBER.len);

    /*  Step 1:  Creating an Algorithm Object */
    if ((status = B_CreateAlgorithmObject (&rsaDecryptor)) != 0)
      break;

    /*  Step 2:  Set the algorithm object to AI_PKCS_RSAPrivate */
    if ((status = B_SetAlgorithmInfo
         (rsaDecryptor, AI_PKCS_RSAPrivate, NULL_PTR)) != 0)
      break;

    /*  Step 3:  Init -- Use the private key associated with the public
                 key used to encrypt */
    if ((status = B_DecryptInit
         (rsaDecryptor, privateKey, RSA_SAMPLE_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Update */
    generalFlag = 0;
    if ((status = B_DecryptUpdate
         (rsaDecryptor, decryptedData, &outputLenUpdate, BLOCK_SIZE,
          encryptedData, outputLenTotal, NULL_PTR,
          &generalSurrenderContext)) != 0)
      break;

    /*  Step 5:  Final */
    generalFlag = 0;
    if ((status = B_DecryptFinal
         (rsaDecryptor, decryptedData + outputLenUpdate,
          &outputLenFinal, BLOCK_SIZE - outputLenUpdate, NULL_PTR,
          &generalSurrenderContext)) != 0)
      break;

    outputLenTotal = outputLenUpdate + outputLenFinal;

    printf ("\nThe decrypted data (%u bytes): \n", outputLenTotal);
    PrintBuf (decryptedData, outputLenTotal);

    if ((status = T_memcmp (dataToEncryptWithRSA, decryptedData,
                            outputLenTotal)) != 0)
      break;

    printf ("The decrypted data matches the original data to encrypt.\n");
    
  } while (0);

  if (status != 0) {
    printf ("Status = %i \n", status);
    printf ("Digital Envelope failed");
  }  

  /*  Step 6:  Destroy
   */
  B_DestroyAlgorithmObject (&randomAlgorithm);
  B_DestroyAlgorithmObject (&keypairGenerator);
  B_DestroyAlgorithmObject (&rsaEncryptor);
  B_DestroyAlgorithmObject (&rsaDecryptor);
  B_DestroyKeyObject (&privateKey);
  B_DestroyKeyObject (&publicKey);

  /*  Free up any memory allocated
   */
  T_free (randomSeed);
  T_free (myPublicKeyBER.data);
  T_free (privateKeyBER.data);

} /*  end main  */

/*   General Surrender Function   */
/*   ==========================   */
int GeneralSurrenderFunction (handle)
POINTER handle;
{
  char s[100];
  static time_t currentTime;
  time_t getTime;

  if ((int)*handle == 0) {
    getTime = time(NULL);
    strftime (s, 100, "%H:%M:%S on %A, %d %B %Y", localtime(&getTime));
    printf ("\n%s\n", s);
    printf ("Surrender function ...\n");
    *handle = 1;
    time (&currentTime);
  }
  else {
    time (&getTime);
    if (currentTime != getTime) {
      printf (" .");
      currentTime = getTime;
    }
  }
  return (0);
} /*  end GeneralSurrenderFunction  */


/*  This procedure will print out what's in the buffer.
 */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */
