/* Copyright (C) RSA Data Security, Inc. created 1995, 1998.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

/*  This program will prompt the user for some input to digest.  You may
      want to rerun this program using the same input data to prove to
      yourself that you get the same digest.  A slight change to the data
      will result in a totally different digest.
    This program is a slight variation to MDIGEST.C in that it illustrates
      the BER encoding of a digest.
 */

#include <stdio.h>   
#include <string.h>  
#include "aglobal.h"
#include "bsafe.h"

#define DIGEST_LEN 20
/* 15 bytes for SHA1, 18 for MD5 */
#define DIGEST_INFO_LEN DIGEST_LEN + 15
#define BUF_LEN 256         /* arbitrary */

B_ALGORITHM_METHOD *DIGEST_CHOOSER[] = {
  &AM_SHA,
  (B_ALGORITHM_METHOD *)NULL_PTR
};

static int GetCommand PROTO_LIST ((char *, unsigned int, char *));

void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));

void main()
{
  B_ALGORITHM_OBJ digester = (B_ALGORITHM_OBJ)NULL_PTR;
  unsigned char dataToDigest[BUF_LEN];
  unsigned int dataToDigestLen;
  unsigned char digestedData[DIGEST_LEN];
  unsigned int digestedDataLen;
                                   
  ITEM *sha1AlgInfoBER;
  /* buffer to hold digest OID information & 20-byte digest */
  unsigned char digestInfoBER[DIGEST_INFO_LEN];  
  unsigned int digestInfoBERLen;

  ITEM retrievedAlgorithmID;
  ITEM retrievedDigest;
  unsigned int status;

  do {
    printf ("BER-encoded Message Digest using SHA1 \n");
    printf ("===================================== \n");

    /*  Step 1:  Create an algorithm object */
    if ((status = B_CreateAlgorithmObject (&digester)) != 0)
      break;

    /*  Step 2:  Set the algorithm object to SHA1 */
    if ((status = B_SetAlgorithmInfo
         (digester, AI_SHA1, NULL_PTR)) != 0)
      break;

    /*  Step 3:  Initialize the digest algorithm. Prompt the user for input.
     */
    if ((status = B_DigestInit
         (digester, (B_KEY_OBJ)NULL_PTR, DIGEST_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    do {
      status =
         GetCommand ((char *)dataToDigest, BUF_LEN,
                     "\nEnter data to digest:");
    } while (status != 0);
   
    dataToDigestLen = (unsigned int)strlen ((char *)dataToDigest);
    printf ("\nLength of input = %i \n", dataToDigestLen);
    printf ("\nData to digest in hex: \n");
    PrintBuf (dataToDigest, dataToDigestLen);

    /*  Step 4:  Update -- Digest the user's input */
    if ((status = B_DigestUpdate
         (digester, dataToDigest, dataToDigestLen,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 5:  Final */
    if ((status = B_DigestFinal
         (digester, digestedData, &digestedDataLen, DIGEST_LEN,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    printf ("Digested data: \n");
    PrintBuf (digestedData, digestedDataLen);

    /*  BER-encode the digest
     */
    if ((status = B_GetAlgorithmInfo
         ((POINTER *)&sha1AlgInfoBER, digester, AI_SHA1_BER)) != 0)
      break;

    if ((status = B_EncodeDigestInfo
         (digestInfoBER, &digestInfoBERLen, DIGEST_INFO_LEN, sha1AlgInfoBER,
          digestedData, digestedDataLen)) != 0)
      break;

    printf ("BER-encoded digest: \n");
    PrintBuf (digestInfoBER, digestInfoBERLen);

    /*  BER-decode the BER-encoded digest
     */
    if ((status = B_DecodeDigestInfo
         (&retrievedAlgorithmID, &retrievedDigest, digestInfoBER,
          digestInfoBERLen)) != 0)
      break;

    printf ("Retrieved Algorithm: \n");
    PrintBuf (retrievedAlgorithmID.data, retrievedAlgorithmID.len);
          
    printf ("Retrieved digest: \n");
    PrintBuf (retrievedDigest.data, retrievedDigest.len);

    if ((status = (digestedDataLen != retrievedDigest.len)) != 0)
      break;

    if ((status = T_memcmp (digestedData, retrievedDigest.data,
                            retrievedDigest.len)) != 0)
      break;

    printf ("Retrieved digest matches with digested data.\n");

  } while (0);

  if (status != 0) {
    printf ("Status = %i \n", status);
    printf ("BER-encoded Message Digest using SHA1 failed");
  }
  
  /*  Step 6:  Remember to destroy all objects, and free up any memory
        allocated
   */
  B_DestroyAlgorithmObject (&digester);

  T_memset (dataToDigest, 0, BUF_LEN);

} /*  end main  */


static int GetCommand (command, maxCommandSize, prompt)
char *command;
unsigned int maxCommandSize;
char *prompt;
{
  unsigned int i;
  

  printf ("%s\n", prompt);
  fflush (stdout);
 
  fgets (command, maxCommandSize, stdin);

  if (*command == '\n' || *command == '\r')
    return (-1);
  
  /* Replace the line terminator with a '\0'.
   */
  for (i = 0; command[i] != '\0'; i++) {
    if (command[i] == '\n' || command[i] == '\r' ||
        i == (maxCommandSize - 1)) {
      command[i] = '\0';
      return (0);
    }
  }
  return (0);
} /*  end GetCommand  */


/*  This procedure will print out what's in the buffer.
 */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */
