/*   Copyright (C) RSA Data Security, Inc. created 1998.
**
**   This file is used to demonstrate how to interface to an
**   RSA Data Security, Inc. licensed development product.
**
**   You have a royalty-free right to use, modify, reproduce and
**   distribute this demonstration file (including any modified
**   version), provided that you agree that RSA Data Security,
**   Inc. has no warranty, implied or otherwise, or liability
**   for this demonstration file or any modified version.
*/

/*  This program will demonstrate the Diffie-Hellman key agreement
 *  scheme where two parties exchange public values to agree upon
 *  a shared secret key.
 *
 *  If you do not wish to use the acceleration table, when compiling
 *  use -DACCEL_TABLE=0, or something similar.
 *
 *  If you want to use the hard-coded EC parameters below, compile with
 *  -DGEN_PARAMS=0, or equivalent.  Otherwise, this program will generate
 *  new EC Parameters using the information in the EC Parameter Generation
 *  Information section.
 */

#include <stdio.h>
#include <time.h>
#include "aglobal.h"
#include "bsafe.h"

#ifndef ACCEL_TABLE
#define ACCEL_TABLE 1
#endif

#ifndef GEN_PARAMS
#define GEN_PARAMS 1
/* Information for EC Parameter Generation  */

/* Currently, the only implementation version supported is 0 */
#define ECPG_VER 0

/* Specify the base field type for the elliptic curve:  FT_FP,
   FT_F2_POLYNOMIAL, or FT_F2_ONB.  See below for restrictions on
   the field element length depending on the field type chosen.  */
#define ECPG_FT FT_FP

/* Specify the length of the prime, basis polynomial, or degree of
   the field -- whichever applies, for fieldElementBits.
   Legal values for field element bits:
     *  with FT_FP or FT_F2_POLYNOMIAL:  64 - 256 bits
     *  with FT_F2_ONB: 65,66,69,74,81,82,83,86,89,90,95,98,99,100,105,
                        106,113,119,130,131,134,135,138,146,148,155,158,
                        162,172,173,174,178,179,180,183,186,189,191,194,
                        196,209,210,221,226,230,231,233,239,243,245,251,254  */
#define ECPG_FE_BITS 160

/* Currently compression is not supported */
#define ECPG_CI CI_NO_COMPRESS

/* Specify the minimum size of the group generated by the base.
   Input of 0 defaults to (fieldElementBits - 7), for minOrderBits */
#define ECPG_MO_BITS 0

/* Specify the maximum size of the second largest prime subgroup of the group.
   Input of 0 defaults to 255, for trialDivBound */
#define ECPG_TDB 0

/* Set if the use of precomputed curves is desired.  This option is only
   available for even polynomial representaions. */
#define ECPG_TL 0
#endif 
/*  end EC Parameter Generation Information  */

int InitializeRandomAlgorithm PROTO_LIST ((B_ALGORITHM_OBJ *));
int InitializeECParamsObj PROTO_LIST ((B_ALGORITHM_OBJ *, B_ALGORITHM_OBJ *));

#if ACCEL_TABLE
int BuildAccelTable PROTO_LIST ((ITEM *, B_ALGORITHM_OBJ *));
#endif

int AllocAndCopyECParamInfo PROTO_LIST ((A_EC_PARAMS *output,
                                      A_EC_PARAMS *input));
void FreeECParamInfo PROTO_LIST ((A_EC_PARAMS *ecParams));
int PrintECParamInfo PROTO_LIST ((B_ALGORITHM_OBJ *));
void PrintBufAsCArray PROTO_LIST ((char *, ITEM *));
void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));
int GeneralSurrenderFunction PROTO_LIST ((POINTER handle));

B_ALGORITHM_METHOD *EC_DH_CHOOSER[] = {
  &AM_ECFP_DH_KEY_AGREE,
  &AM_ECF2POLY_DH_KEY_AGREE,
  (B_ALGORITHM_METHOD *)NULL_PTR
};

void main ()
{
  int status;
  
  B_ALGORITHM_OBJ randomAlgorithm = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ alice = (B_ALGORITHM_OBJ)NULL_PTR;
  B_ALGORITHM_OBJ bob = (B_ALGORITHM_OBJ)NULL_PTR;  
  B_ALGORITHM_OBJ ecParamsObj = (B_ALGORITHM_OBJ)NULL_PTR;
  B_EC_PARAMS commonECParams;
  A_EC_PARAMS *cryptocECParams;
  A_EC_PARAMS ecParams;  

  unsigned int fieldElementLen, maxPhase1Len, maxPhase2Len;
  unsigned char *alicePublicValue = NULL_PTR;
  unsigned char *bobPublicValue = NULL_PTR;
  unsigned int alicePublicValueLen, bobPublicValueLen;
  unsigned char *aliceSecretValue = NULL_PTR;
  unsigned char *bobSecretValue = NULL_PTR;
  unsigned int aliceSecretValueLen, bobSecretValueLen;

#if ACCEL_TABLE
  ITEM aTableItem;
#endif
  
  do {
    printf ("EC Diffie-Hellman Key Agreement Algorithm \n");
    printf ("========================================= \n");

    if ((status = InitializeRandomAlgorithm(&randomAlgorithm)) != 0)
      break;
    if ((status = InitializeECParamsObj(&ecParamsObj, &randomAlgorithm)) != 0)
      break;
    if ((status = PrintECParamInfo(&ecParamsObj)) != 0)
      break;
#if ACCEL_TABLE
    if ((status = BuildAccelTable (&aTableItem, &ecParamsObj)) != 0)
      break;
#endif
    
    /*  Step 1:  Create an algorithm object  */
    if ((status = B_CreateAlgorithmObject(&alice)) != 0)
      break;
    if ((status = B_CreateAlgorithmObject(&bob)) != 0)
      break;
    
    /*  Step 2:  Set algorithm object to AI_EC_DHKeyAgree  */
    /*  First get the EC parameters that Alice and Bob will have in common */
    if ((status = B_GetAlgorithmInfo((POINTER *)&cryptocECParams, ecParamsObj,
                                     AI_ECParameters)) != 0)
      break;

    if ((status = AllocAndCopyECParamInfo(&ecParams, cryptocECParams)) != 0)
      break;    
    
    commonECParams.parameterInfoType = AI_ECParameters;
    commonECParams.parameterInfoValue = (POINTER)&ecParams;
    /*  Provide Alice and Bob with the same EC parameters  */
    if ((status = B_SetAlgorithmInfo(alice, AI_EC_DHKeyAgree,
                                     (POINTER)&commonECParams)) != 0)
      break;
    if ((status = B_SetAlgorithmInfo(bob, AI_EC_DHKeyAgree,
                                     (POINTER)&commonECParams)) != 0)
      break;

#if ACCEL_TABLE
    /*  Alice and Bob can also optionally receive the acceleration tables
     *  corresponding with the EC parameters they have.
     */
    if ((status = B_SetAlgorithmInfo (alice, AI_ECAcceleratorTable,
                                      (POINTER)&aTableItem)) != 0)
      break;
    if ((status = B_SetAlgorithmInfo (bob, AI_ECAcceleratorTable,
                                      (POINTER)&aTableItem)) != 0)
      break;

    T_memset (aTableItem.data, 0, aTableItem.len);
#endif

    /*  Step 3:  Initialize key agreement protocol  */
    if ((status = B_KeyAgreeInit(alice, (B_KEY_OBJ)NULL_PTR, EC_DH_CHOOSER,
                                 (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
    if ((status = B_KeyAgreeInit(bob, (B_KEY_OBJ)NULL_PTR, EC_DH_CHOOSER,
                                 (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  The largest size of phase 1 output we can get is one byte larger
     *  than twice the field element size.  For phase 2, the size of the
     *  output should be the same as the field element size.  (See
     *  Library Reference Manual entry on AI_EC_DHKeyAgree for more details)
     */
    fieldElementLen = (ecParams.fieldElementBits + 7) / 8;
    printf("EC Parameters:  Field element length = %u bytes\n",
           fieldElementLen);
    
    maxPhase1Len = (fieldElementLen * 2) + 1;
    maxPhase2Len = fieldElementLen;
    
    /*  Step 4:  Phase 1  */
    printf ("\nKey Agreement -- Phase 1 \n");
    printf ("======================== \n");
    
    alicePublicValue = T_malloc(maxPhase1Len);
    if ((status = (alicePublicValue == NULL_PTR)) != 0)
      break;
    bobPublicValue = T_malloc(maxPhase1Len);
    if ((status = (bobPublicValue == NULL_PTR)) != 0)
      break;

    if ((status = B_KeyAgreePhase1(alice, alicePublicValue,
                                   &alicePublicValueLen, maxPhase1Len,
                                   randomAlgorithm,
                                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
    printf("\nAlice's public value (%u bytes):\n", alicePublicValueLen);
    PrintBuf(alicePublicValue, alicePublicValueLen);
    
    if ((status = B_KeyAgreePhase1(bob, bobPublicValue,
                                   &bobPublicValueLen, maxPhase1Len,
                                   randomAlgorithm,
                                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
    printf("\nBob's public value (%u bytes):\n", bobPublicValueLen);
    PrintBuf(bobPublicValue, bobPublicValueLen);

    /*  Step 5:  Phase 2  */
    printf ("\nKey Agreement -- Phase 2 \n");
    printf ("======================== \n");
    
    aliceSecretValue = T_malloc(maxPhase2Len);
    if ((status = (aliceSecretValue == NULL_PTR)) != 0)
      break;
    bobSecretValue = T_malloc(maxPhase2Len);
    if ((status = (bobSecretValue == NULL_PTR)) != 0)
      break;
    
    if ((status = B_KeyAgreePhase2(alice, aliceSecretValue,
                                   &aliceSecretValueLen, maxPhase2Len,
                                   bobPublicValue, bobPublicValueLen,
                                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
    printf("\nValue derived by Alice (%u bytes):\n", aliceSecretValueLen);
    PrintBuf(aliceSecretValue, aliceSecretValueLen);

    if ((status = B_KeyAgreePhase2(bob, bobSecretValue,
                                   &bobSecretValueLen, maxPhase2Len,
                                   alicePublicValue, alicePublicValueLen,
                                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
    printf("\nValue derived by Bob (%u bytes):\n", bobSecretValueLen);
    PrintBuf(bobSecretValue, bobSecretValueLen);

    /*  Alice and Bob should derive the same secret value  */
    if ((status = (aliceSecretValueLen != bobSecretValueLen)) != 0)
      break;
    if ((status = T_memcmp(aliceSecretValue, bobSecretValue,
                           aliceSecretValueLen)) != 0)
      break;
    printf("Alice and Bob's secret keys agree\n");
  } while(0);

  if (status != 0) 
    printf ("Key agreement failed with status = %i\n", status);

#if ACCEL_TABLE
  T_free(aTableItem.data);
#endif
  FreeECParamInfo(&ecParams);
  T_free (alicePublicValue);
  T_free (bobPublicValue);
  T_free (aliceSecretValue);
  T_free (bobSecretValue);
  B_DestroyAlgorithmObject(&randomAlgorithm);
  B_DestroyAlgorithmObject(&alice);
  B_DestroyAlgorithmObject(&bob);
}  /* end main */

/*  Create a random algorithm
 *  Pass in a pointer to a B_ALGORITHM_OBJ and be sure to call
 *  B_DestroyAlgorithmObject later in the function that calls
 *  this one.
 */
int InitializeRandomAlgorithm (randomAlgorithm)
B_ALGORITHM_OBJ* randomAlgorithm;
{
  int status;

  POINTER randomSeed = NULL_PTR;
  unsigned int randomSeedLen;

  static B_ALGORITHM_METHOD *randomChooser[] = {
    &AM_SHA_RANDOM,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  /*  The following code was pulled from genbytes.c, steps 1-4.  */
  do {
    if ((status = B_CreateAlgorithmObject (randomAlgorithm)) != 0)
      break;

    if ((status = B_SetAlgorithmInfo
                  (*randomAlgorithm, AI_SHA1Random, NULL_PTR)) != 0)
      break;

    if ((status = B_RandomInit
                  (*randomAlgorithm, randomChooser,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    randomSeedLen = 88;
    randomSeed = T_malloc (randomSeedLen);
    if ((status = (randomSeed == NULL_PTR)) != 0)
      break;

    T_memset (randomSeed, 0, randomSeedLen);

    puts ("Enter a random seed");
    if ((status =
	 (NULL_PTR == (unsigned char *)gets ((char *)randomSeed))) != 0)
      break;

    printf("The random seed in hex (%u bytes):\n", randomSeedLen);
    PrintBuf(randomSeed, randomSeedLen);

    if ((status = B_RandomUpdate
                  (*randomAlgorithm, randomSeed, randomSeedLen,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
  } while(0);
  
  T_free(randomSeed);

  if (status != 0)
    printf ("Failed to initialize random algorithm.  Status = %i\n", status);
  
  return status;
}    /*  end InitializeRandomAlgorithm   */

/*  Initializes the algorithm object with a set of EC parameters.
 *  Be sure to call B_DestroyAlgorithmObject on ecParamObj after
 *  the do-while in the function that calls this one.
 */
int InitializeECParamsObj (ecParamsObj, randomAlgorithm)
B_ALGORITHM_OBJ *ecParamsObj;
B_ALGORITHM_OBJ *randomAlgorithm;
{
  int status;

#if !GEN_PARAMS
  A_EC_PARAMS stockECParams;

  unsigned char fieldInfo[20] = {
    0xDA, 0xE2, 0x12, 0xCC, 0xEC, 0x6D, 0xFA, 0x73,
    0x17, 0x44, 0x1C, 0xEE, 0x28, 0xF0, 0x42, 0xA3,
    0xDE, 0xD0, 0x4D, 0x7F
  };

  unsigned char coeffA[20] = {
    0xDA, 0xE2, 0x12, 0xCC, 0xEC, 0x6D, 0xFA, 0x73,
    0x17, 0x44, 0x1C, 0xEE, 0x28, 0xF0, 0x42, 0xA3,
    0xDE, 0xD0, 0x4D, 0x7C
  };

  unsigned char coeffB[20] = {
    0xBF, 0x63, 0x40, 0xB3, 0xF8, 0xEF, 0x6A, 0xBC,
    0xD1, 0x9B, 0x56, 0x37, 0x69, 0x85, 0x5B, 0xA0,
    0xA2, 0xAE, 0x84, 0x92
  };

  unsigned char base[41] = {
    0x04, 0x77, 0x79, 0xDC, 0x0B, 0xF7, 0xFA, 0x7E,
    0x52, 0xD1, 0x4C, 0x14, 0x3A, 0x60, 0x7A, 0x46,
    0xE3, 0x6C, 0x7B, 0x7A, 0x7E, 0xD1, 0xA0, 0xC5,
    0x30, 0xA6, 0x2B, 0xF5, 0x4F, 0xA8, 0xE7, 0x6F,
    0x58, 0x64, 0xCC, 0x5A, 0xF3, 0xAB, 0x06, 0x76,
    0x6A
  };

  unsigned char order[20] = {
    0x06, 0x14, 0x80, 0x85, 0xB1, 0x3B, 0xF1, 0x9F,
    0xA4, 0x33, 0xA9, 0x32, 0x42, 0x85, 0x00, 0xFF,
    0x30, 0x43, 0x2E, 0x75
  };

  unsigned char cofactor[1] = {
    0x24
  };

#else
  static B_ALGORITHM_METHOD *paramGenChooser[] = {
    &AM_ECFP_PARAM_GEN,
    &AM_ECF2POLY_PARAM_GEN,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };
  
  B_ALGORITHM_OBJ paramGenObj = (B_ALGORITHM_OBJ)NULL_PTR;
  B_EC_PARAM_GEN_PARAMS paramGenInfo;

  A_SURRENDER_CTX generalSurrenderContext;
  int generalFlag;
#endif
  
  do {
    /*  Step 1:  Create algorithm object to hold EC Parameters */
    if ((status = B_CreateAlgorithmObject(ecParamsObj)) != 0)
      break;

#if !GEN_PARAMS
    /*  Step 2:  Set with old hard-coded EC Parameters */
    stockECParams.version = 0x00;
    stockECParams.fieldType = FT_FP;
    stockECParams.compressIndicator = CI_NO_COMPRESS;
    stockECParams.fieldElementBits = 160;
    stockECParams.fieldInfo.data = fieldInfo;
    stockECParams.fieldInfo.len = sizeof(fieldInfo);
    stockECParams.coeffA.data = coeffA;
    stockECParams.coeffA.len = sizeof(coeffA);
    stockECParams.coeffB.data = coeffB;
    stockECParams.coeffB.len = sizeof(coeffB);
    stockECParams.base.data = base;
    stockECParams.base.len = sizeof(base);
    stockECParams.order.data = order;
    stockECParams.order.len = sizeof(order);
    stockECParams.cofactor.data = cofactor;
    stockECParams.cofactor.len = sizeof(cofactor);
    
    if ((status = B_SetAlgorithmInfo (*ecParamsObj, AI_ECParameters,
                                      (POINTER)&stockECParams)) != 0)
      break;
#else
    /*  Step 1:  Create algorithm object to generate new parameters  */
    if ((status = B_CreateAlgorithmObject(&paramGenObj)) != 0)
      break;

    /*  Step 2:  Set information for generating new EC parameters  */
    paramGenInfo.version = ECPG_VER;
    paramGenInfo.fieldType = ECPG_FT;  
    paramGenInfo.fieldElementBits = ECPG_FE_BITS; 
    paramGenInfo.compressIndicator = ECPG_CI;
    paramGenInfo.minOrderBits = ECPG_MO_BITS;
    paramGenInfo.trialDivBound = ECPG_TDB;
    paramGenInfo.tableLookup = ECPG_TL;
    if ((status = B_SetAlgorithmInfo(paramGenObj, AI_ECParamGen,
                                     (POINTER)&paramGenInfo)) != 0)
      break;

    /*  Step 3:  Init  */
    if ((status = B_GenerateInit(paramGenObj, paramGenChooser,
                                 (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Generate  */
    printf ("\nGenerating new EC Parameters ");
    printf ("(this could take some time)...\n");
    generalSurrenderContext.Surrender = GeneralSurrenderFunction;
    generalSurrenderContext.handle = (POINTER)&generalFlag;
    generalSurrenderContext.reserved = NULL_PTR;
    generalFlag = 0;
    if ((status = B_GenerateParameters(paramGenObj, *ecParamsObj,
                                       *randomAlgorithm,
                                       &generalSurrenderContext)) != 0)
      break;
#endif
  } while(0);

  if (status != 0)
    printf("Failed to generate new EC parameters.  Status = %i\n", status);

#if GEN_PARAMS
  B_DestroyAlgorithmObject (&paramGenObj);
#endif
  
  return status;
}    /*  end InitializeECParamsObj   */

#if ACCEL_TABLE
/*  Build the appropriate acceleration table for the given EC parameters.
 *  accelTableItem.data will be given a pointer to the location of the
 *  allocated space for the acceleration table.  This memory must be freed
 *  after the do-while in the calling function.
 */
int BuildAccelTable (accelTableItem, ecParamsObj)
ITEM *accelTableItem;
B_ALGORITHM_OBJ *ecParamsObj;
{
  int status;

  static B_ALGORITHM_METHOD *ecAccelChooser[] = {
    &AM_ECFP_BLD_ACCEL_TABLE,
    &AM_ECF2POLY_BLD_ACCEL_TABLE,    
    (B_ALGORITHM_METHOD *)NULL_PTR
  };
  
  B_ALGORITHM_OBJ buildTable = (B_ALGORITHM_OBJ)NULL_PTR;
  B_EC_PARAMS paramInfo;
  A_EC_PARAMS *cryptocECParamInfo;
  A_EC_PARAMS ecParamInfo;
  unsigned int maxTableLen;

  A_SURRENDER_CTX generalSurrenderContext;
  int generalFlag;

  do {
    printf ("\nGenerating Acceleration Table (this may take a while)...\n");
 
    /*  Step 1:  Create  */
    if ((status = B_CreateAlgorithmObject(&buildTable)) != 0)
      break;

    /*  Step 2:  Set  */
    /*  Retrieve the EC parameters and place in the algorithm object
        used to build the acceleration table for those parameters  */
    if ((status = B_GetAlgorithmInfo((POINTER *)&cryptocECParamInfo,
                                     *ecParamsObj, AI_ECParameters)) != 0)
      break;

    if ((status = AllocAndCopyECParamInfo(&ecParamInfo,
                                          cryptocECParamInfo)) != 0)
      break;
    
    paramInfo.parameterInfoType = AI_ECParameters;
    paramInfo.parameterInfoValue = (POINTER)&ecParamInfo;
    
    if ((status = B_SetAlgorithmInfo(buildTable, AI_ECBuildAcceleratorTable,
                                     (POINTER)&paramInfo)) != 0)
      break;

    /*  Step 3:  Init  */
    if ((status = B_BuildTableInit(buildTable, ecAccelChooser,
                                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 5:  Final */
    /*  Allocate sufficient memory to hold the acceleration table  */
    if ((status = B_BuildTableGetBufSize(buildTable, &maxTableLen)) != 0)
      break;
    printf("Max Len of Acceleration Table = %u bytes\n", maxTableLen);
    accelTableItem->data = T_malloc(maxTableLen);
    if ((status = (accelTableItem->data == NULL_PTR)) != 0)
      break;
    
    /*  Build acceleration table  */
    generalSurrenderContext.Surrender = GeneralSurrenderFunction;
    generalSurrenderContext.handle = (POINTER)&generalFlag;
    generalSurrenderContext.reserved = NULL_PTR;
    generalFlag = 0;
    if ((status = B_BuildTableFinal(buildTable, accelTableItem->data,
                                    &(accelTableItem->len), maxTableLen,
                                    &generalSurrenderContext)) != 0)
      break;

    printf("\nAcceleration Table Length = %u bytes\n", accelTableItem->len);
  } while(0);

  if (status != 0)
    printf("Failed to generate acceleration table.  Status = %i\n", status);

  B_DestroyAlgorithmObject(&buildTable);
  FreeECParamInfo(&ecParamInfo);
  
  return status;
}    /*  end BuildAccelTable   */
#endif

/*  This procedure allocates and copies information from input, a pointer
 *  to an A_EC_PARAMS structure containing memory belonging to CRYPTO-C, to
 *  output, a pointer to an A_EC_PARAMS structure created by the application.
 *  Be sure to call FreeECParamInfo with output when appropriate.
 */
int AllocAndCopyECParamInfo(output, input)
A_EC_PARAMS *output;
A_EC_PARAMS *input;
{
  int status;

  do {
    output->version = input->version;
    
    output->fieldType = input->fieldType;

    output->fieldInfo.len = input->fieldInfo.len;
    output->fieldInfo.data = T_malloc(output->fieldInfo.len);
    if ((status = (output->fieldInfo.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->fieldInfo.data, input->fieldInfo.data,
             output->fieldInfo.len);

    output->coeffA.len = input->coeffA.len;
    output->coeffA.data = T_malloc(output->coeffA.len);
    if ((status = (output->coeffA.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->coeffA.data, input->coeffA.data,
             output->coeffA.len);

    output->coeffB.len = input->coeffB.len;
    output->coeffB.data = T_malloc(output->coeffB.len);
    if ((status = (output->coeffB.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->coeffB.data, input->coeffB.data,
             output->coeffB.len);

    output->base.len = input->base.len;
    output->base.data = T_malloc(output->base.len);
    if ((status = (output->base.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->base.data, input->base.data,
             output->base.len);

    output->order.len = input->order.len;
    output->order.data = T_malloc(output->order.len);
    if ((status = (output->order.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->order.data, input->order.data,
             output->order.len);

    output->cofactor.len = input->cofactor.len;
    output->cofactor.data = T_malloc(output->cofactor.len);
    if ((status = (output->cofactor.data == NULL_PTR)) != 0)
      break;
    T_memcpy(output->cofactor.data, input->cofactor.data,
             output->cofactor.len);
    
    output->compressIndicator = input->compressIndicator;
    
    output->fieldElementBits = input->fieldElementBits;
  } while (0);

  if (status != 0)
    printf("AllocAndCopyECParamInfo failed with status %i\n", status);
  
  return status;
}    /*  end AllocAndCopyECParamInfo  */

/*  This procedure takes a pointer to an A_EC_PARAMS structure containing
 *  space allocated by AllocAndCopyECParamInfo and frees all data allocated
 *  with T_malloc.
 */
void FreeECParamInfo(ecParams)
A_EC_PARAMS *ecParams;
{
  T_free(ecParams->fieldInfo.data);
  T_free(ecParams->coeffA.data);
  T_free(ecParams->coeffB.data);
  T_free(ecParams->base.data);
  T_free(ecParams->order.data);
  T_free(ecParams->cofactor.data);
}    /*  end FreeECParamInfo  */

/*  Given an algorithm object containing EC parameters,
 *  print the information out
 */
int PrintECParamInfo(ecParamsObj)
B_ALGORITHM_OBJ *ecParamsObj;
{
  int status;
  A_EC_PARAMS *cryptocECParamInfo;

  do {
    if ((status = B_GetAlgorithmInfo((POINTER *)&cryptocECParamInfo,
                                     *ecParamsObj, AI_ECParameters)) != 0)
      break;
    printf("\n***EC Parameters Generated:\nversion:  %u\n",
           cryptocECParamInfo->version);
    printf("compress indicator:  ");
    switch (cryptocECParamInfo->compressIndicator) {
      case CI_NO_COMPRESS:
        printf("CI_NO_COMPRESS\n");
        break;
      default:
        printf("Unrecognized compression indicator value:  %i\n",
               cryptocECParamInfo->compressIndicator);
        break;
    }    
    printf("field type:  ");
    switch (cryptocECParamInfo->fieldType) {
      case FT_FP:
        printf("FT_FP\n");
        break;
      case FT_F2_ONB:
        printf("FT_F2_ONB\n");
        break;
      case FT_F2_POLYNOMIAL:
        printf("FT_F2_POLYNOMIAL\n");
        break;
      default:
        printf("Unrecognized field type:  %i\n", 
               cryptocECParamInfo->fieldType);
        break;
    }
    printf("field element length:  %u bits\n",
           cryptocECParamInfo->fieldElementBits);
    PrintBufAsCArray("fieldInfo", &cryptocECParamInfo->fieldInfo);
    PrintBufAsCArray("coeffA", &cryptocECParamInfo->coeffA);
    PrintBufAsCArray("coeffB", &cryptocECParamInfo->coeffB);
    PrintBufAsCArray("base", &cryptocECParamInfo->base);
    PrintBufAsCArray("order", &cryptocECParamInfo->order);
    PrintBufAsCArray("cofactor", &cryptocECParamInfo->cofactor);
  } while(0);

  return status;
}    /*  end PrintECParamInfo  */

/*  This procedure will print out what's in the buffer as a C array.
 */
void PrintBufAsCArray(label, buffer)
char *label;
ITEM *buffer ;
{
  unsigned int count;

  printf("unsigned char %s[%u] = {\n", label, buffer->len);
  for (count = 0 ; count < buffer->len; count++) {
    if ((count % 8) == 0) {
      printf("  0x%02X", (buffer->data)[count]);
    } else {
      printf(" 0x%02X", (buffer->data)[count]);
    }
    if (count != (buffer->len - 1)) {
      printf(",");
    }
    if (((count % 8) == 7 ) || (count == (buffer->len - 1))) {
      printf("\n");
    }
  }
  printf("};\n");
} /*  end PrintBufAsCArray  */

/*  This procedure will print out what's in the buffer.
 */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */

int GeneralSurrenderFunction (handle)
POINTER handle;
{
  char s[100];
  static time_t currentTime;
  time_t getTime;

  if ((int)*handle == 0) {
    getTime = time(NULL);
    strftime (s, 100, "%H:%M:%S on %A, %d %B %Y", localtime(&getTime));
    printf ("\n%s\n", s);
    printf ("Surrender function ...\n");
    *handle = 1;
    time (&currentTime);
  }
  else {
    time (&getTime);
    if (currentTime != getTime) {
      printf (" .");
      currentTime = getTime;
    }
  }
  return (0);
} /*  end GeneralSurrenderFunction  */
