/* Copyright (C) RSA Data Security, Inc. created 1995, 1998.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

/*  This program will generate a RSA keypair, encrypt with the public key,
      then decrypt with the private key -- i.e. Digital Envelope.

    Modified to use the SET OAEP form of digital envelopes.
 */

#include <stdio.h>   
#include <time.h>
#include "aglobal.h"
#include "bsafe.h"

/*  Set an arbitrary value for the number of bytes to encrypt  */
#define DATA_TO_ENCRYPT_LEN 8

/*  RSA_MODULUS_BITS must be evenly divisible by 8 for these AI's  */
#define RSA_MODULUS_BITS 1024
#define BLOCK_SIZE ((RSA_MODULUS_BITS + 7) / 8)

/*  For SET OAEP, the padding scheme needs exactly 25 bytes to work with.
    In contrast, PKCS #1 v1.5 padding needs a minimum of 11 bytes.  */
#define BLOCK_OVERHEAD  25

/*  This indicates the number of actual bytes which are available to hold
    the data to encrypt.  In the SET v1.0 spec book 3, this would correspond
    to the ADB, actual data block, on p. 17.  The extra '- 1' is because
    CRYPTO-C expects the first byte to indicate the block contents, BC,
    as specified on p. 16 in book 3 of the SET 1.0 spec.  However, there
    is no specific requirement imposed by CRYPTO-C to adhere to the
    recommendation in the SET spec.  Therefore, AI_SET_OAEP_RSAPublic/Private
    can also be used like raw RSA, except that the size of the data to
    encrypt must be a multiple of (BLOCK_SIZE - 25).  However for the
    purposes of this example, as shown in RSA_Encrypt/Decrypt below,
    we use BC = 0.  */
#define BLOCK_PAYLOAD (BLOCK_SIZE - BLOCK_OVERHEAD - 1)
#define NUM_BLOCKS ((DATA_TO_ENCRYPT_LEN + BLOCK_PAYLOAD - 1) / BLOCK_PAYLOAD)
   
int InitializeRandomAlgorithm PROTO_LIST ((B_ALGORITHM_OBJ *));
int GenerateRSAKeypair PROTO_LIST ((B_ALGORITHM_OBJ *, B_KEY_OBJ *,
                                    B_KEY_OBJ *));
int RSAEncrypt PROTO_LIST ((B_ALGORITHM_OBJ *, B_KEY_OBJ *,
                            ITEM *, ITEM *));
int RSADecrypt PROTO_LIST ((B_KEY_OBJ *, ITEM *, ITEM *));
int GeneralSurrenderFunction PROTO_LIST ((POINTER handle));
void PrintBuf PROTO_LIST ((unsigned char *, unsigned int));

void main()
{
  B_ALGORITHM_OBJ randomAlgorithm = (B_ALGORITHM_OBJ)NULL_PTR;
  B_KEY_OBJ publicKey = (B_KEY_OBJ)NULL_PTR;
  B_KEY_OBJ privateKey = (B_KEY_OBJ)NULL_PTR;

  ITEM dataToEncrypt, encryptedData, decryptedData;
  
  unsigned char dataToEncryptWithRSA[DATA_TO_ENCRYPT_LEN];

  unsigned int status;

  do {
    printf ("\nRSA With SET OAEP Algorithm \n");
    printf ("=========================== \n");

    if ((status = InitializeRandomAlgorithm (&randomAlgorithm)) != 0)
      break;

    if ((status = B_GenerateRandomBytes
                  (randomAlgorithm, dataToEncryptWithRSA, DATA_TO_ENCRYPT_LEN,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    dataToEncrypt.data = dataToEncryptWithRSA;
    dataToEncrypt.len = DATA_TO_ENCRYPT_LEN;
    printf ("Data to Encrypt (%u bytes):\n", dataToEncrypt.len);
    PrintBuf (dataToEncrypt.data, dataToEncrypt.len);

    if ((status = GenerateRSAKeypair (&randomAlgorithm, &publicKey,
                                      &privateKey)) != 0)
      break;

    encryptedData.len = BLOCK_SIZE * NUM_BLOCKS;
    encryptedData.data = T_malloc(encryptedData.len);
    if (encryptedData.data == NULL_PTR) {
      status = BE_ALLOC;
      break;
    }

    if ((status = RSAEncrypt (&randomAlgorithm, &publicKey, &dataToEncrypt,
                              &encryptedData)) != 0)
      break;

    decryptedData.data = T_malloc(BLOCK_PAYLOAD * NUM_BLOCKS);
    if (decryptedData.data == NULL_PTR) {
      status = BE_ALLOC;
      break;
    }    
    if ((status = RSADecrypt (&privateKey, &encryptedData,
                              &decryptedData)) != 0)
      break;
        
    if ((status = T_memcmp (dataToEncrypt.data, decryptedData.data,
                            decryptedData.len)) != 0)
      break;

    printf ("The decrypted data matches the original data to encrypt.\n");
    
  } while (0);

  if (status != 0) {
    printf ("Status = %i \n", status);
    printf ("Digital Envelope failed");
  }  

  /*  Step 6:  Destroy */
  B_DestroyAlgorithmObject (&randomAlgorithm);
  B_DestroyKeyObject (&privateKey);
  B_DestroyKeyObject (&publicKey);
  T_free (encryptedData.data);
  T_free (decryptedData.data);
} /*  end main  */

/*  Create a random algorithm
 *  Pass in a pointer to a B_ALGORITHM_OBJ and be sure to call
 *  B_DestroyAlgorithmObject later in the function that calls
 *  this one.
 */
int InitializeRandomAlgorithm (randomAlgorithm)
B_ALGORITHM_OBJ* randomAlgorithm;
{
  int status;

  POINTER randomSeed = NULL_PTR;
  unsigned int randomSeedLen;

  static B_ALGORITHM_METHOD *randomChooser[] = {
    &AM_SHA_RANDOM,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  /*  The following code was pulled from genbytes.c, steps 1-4.  */
  do {
    if ((status = B_CreateAlgorithmObject (randomAlgorithm)) != 0)
      break;

    if ((status = B_SetAlgorithmInfo
                  (*randomAlgorithm, AI_SHA1Random, NULL_PTR)) != 0)
      break;

    if ((status = B_RandomInit
                  (*randomAlgorithm, randomChooser,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    randomSeedLen = 80;
    randomSeed = T_malloc (randomSeedLen);
    if ((status = (randomSeed == NULL_PTR)) != 0)
      break;

    T_memset (randomSeed, 0, randomSeedLen);

    puts ("Enter a random seed");
    if ((status =
	 (NULL_PTR == (unsigned char *)gets ((char *)randomSeed))) != 0)
      break;

    printf("The random seed in hex (%u bytes):\n", randomSeedLen);
    PrintBuf(randomSeed, randomSeedLen);

    if ((status = B_RandomUpdate
                  (*randomAlgorithm, randomSeed, randomSeedLen,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
  } while(0);
  
  T_free(randomSeed);

  if (status != 0)
    printf ("Failed to initialize random algorithm.  Status = %i\n", status);
  
  return status;
}    /*  end InitializeRandomAlgorithm   */

/*  Create an RSA public/private keypair
 *  randomAlgorithm is a pointer to an already-initialized random algorithm
 *  object.  Be sure to call B_DestroyKeyObject on the public and private
 *  keys.
 */
int GenerateRSAKeypair (randomAlgorithm, publicKey, privateKey)
B_ALGORITHM_OBJ *randomAlgorithm;
B_KEY_OBJ *publicKey;
B_KEY_OBJ *privateKey;
{
  static B_ALGORITHM_METHOD *RSA_KEYGEN_CHOOSER[] = {
    &AM_RSA_KEY_GEN,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  B_ALGORITHM_OBJ keypairGenerator = (B_ALGORITHM_OBJ)NULL_PTR;
  A_RSA_KEY_GEN_PARAMS keygenParams;

  static unsigned char f4Data[] = {0x01, 0x00, 0x01};

  A_SURRENDER_CTX generalSurrenderContext;
  int generalFlag, status;

  generalSurrenderContext.Surrender = GeneralSurrenderFunction;
  generalSurrenderContext.handle = (POINTER)&generalFlag;
  generalSurrenderContext.reserved = NULL_PTR;  

  do {
    printf ("   Generating a Keypair \n");
    printf ("   ====================");

    /*  Step 1:  Create algorithm object */
    if ((status = B_CreateAlgorithmObject (&keypairGenerator)) != 0)
      break;

    /*  Step 2:  Set algorithm object to AI_RSAKeyGen
     */
    keygenParams.modulusBits = RSA_MODULUS_BITS;
    keygenParams.publicExponent.data = f4Data;
    keygenParams.publicExponent.len = sizeof(f4Data);
    if ((status = B_SetAlgorithmInfo (keypairGenerator, AI_RSAKeyGen,
                                      (POINTER)&keygenParams)) != 0)
      break;

    /*  Step 3:  Init */
    if ((status = B_GenerateInit
                  (keypairGenerator, RSA_KEYGEN_CHOOSER,
                   (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  no step 4 (Update) in generating a keypair */

    /*  Step 5:  Generate
     */
    if ((status = B_CreateKeyObject (publicKey)) != 0)
      break;
    if ((status = B_CreateKeyObject (privateKey)) != 0)
      break;

    /*  generalFlag is for the surrender function */
    generalFlag = 0;
    if ((status = B_GenerateKeypair
                  (keypairGenerator, *publicKey, *privateKey,
                   *randomAlgorithm, &generalSurrenderContext)) != 0)
      break;

  } while(0);

  if (status != 0)
    printf ("Failed to generate RSA keypair.  Status = %i\n", status);

  B_DestroyAlgorithmObject (&keypairGenerator);

  return status;
}     /*  end GenerateRSAKeypair  */

/*  Supports RSA public-key encryption using OAEP.
 *  dataToEncrypt.data must point to a block of memory large enough to
 *  hold the encrypted data (a multiple of the modulus size)
 */
int RSAEncrypt (randomAlgorithm, publicKey, dataToEncrypt, encryptedData)
B_ALGORITHM_OBJ *randomAlgorithm;
B_KEY_OBJ *publicKey;
ITEM *dataToEncrypt;
ITEM *encryptedData;
{
  static B_ALGORITHM_METHOD *ENCRYPTION_CHOOSER[] = {
    &AM_RSA_ENCRYPT,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };
  
  B_ALGORITHM_OBJ rsaEncryptor = (B_ALGORITHM_OBJ)NULL_PTR;

  unsigned int inputDataLeft;
  
  unsigned char inputBlock[BLOCK_PAYLOAD + 1];  /*  + 1 for BC  */
  unsigned int inputBlockIndex, encryptedDataIndex;
  unsigned int status, outputLenUpdate, outputLenFinal;
  
  do {
    printf ("\n   Encrypting with the RSA Public Key \n");
    printf ("   ================================== \n");

    /*  Step 1:  Creating an Algorithm Object */
    if ((status = B_CreateAlgorithmObject (&rsaEncryptor)) != 0)
      break;

    /*  Step 2:  Set the algorithm object to AI_SET_OAEP_RSAPublic */
    if ((status = B_SetAlgorithmInfo
         (rsaEncryptor, AI_SET_OAEP_RSAPublic, NULL_PTR)) != 0)
      break;

    /*  Step 3:  Init -- encrypt with the recipient's public key */
    if ((status = B_EncryptInit
         (rsaEncryptor, *publicKey, ENCRYPTION_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Update */
    /*  inputBlockIndex is used to tell us how far into the
        dataToEncrypt->data buffer we've gone  */
    inputBlockIndex = 0;
    
    /*  encryptedDataIndex tells us where to begin writing to
        encryptedData->data  */
    encryptedDataIndex = 0;

    /*  number of bytes remaining in dataToEncrypt->data  */
    inputDataLeft = dataToEncrypt->len;
    
    while (inputDataLeft > 0) {
      /*  zeroize the inputBlock array and write data to encrypt beginning
          at inputBlock[1], so that the first byte will be zero and any
          unused space at the end of inputData will be zero.  SET v1.0
          book 3, pp.16-17 where BC=0.  */
      T_memset (inputBlock, 0, BLOCK_PAYLOAD + 1);

      /*  Write as many bytes as possible from dataToEncrypt->data to
          inputBlock.  */
      if (inputDataLeft < BLOCK_PAYLOAD) {
        T_memcpy (inputBlock + 1, dataToEncrypt->data + inputBlockIndex,
                  inputDataLeft);
        inputDataLeft = 0;
      }
      else {
        T_memcpy (inputBlock + 1, dataToEncrypt->data + inputBlockIndex,
                  BLOCK_PAYLOAD);
        inputDataLeft -= BLOCK_PAYLOAD;
        inputBlockIndex += BLOCK_PAYLOAD;
      }
      
      if ((status = B_EncryptUpdate
                    (rsaEncryptor, encryptedData->data + encryptedDataIndex,
                     &outputLenUpdate, BLOCK_SIZE, (unsigned char *)inputBlock,
                     BLOCK_PAYLOAD + 1, *randomAlgorithm,
                     (A_SURRENDER_CTX *)NULL_PTR)) != 0)
        break;

      /*  Step 5:  Final */
      if ((status = B_EncryptFinal
            (rsaEncryptor,
             encryptedData->data + encryptedDataIndex + outputLenUpdate,
             &outputLenFinal, BLOCK_SIZE - outputLenUpdate,
             *randomAlgorithm, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
        break;
      encryptedDataIndex += outputLenUpdate + outputLenFinal;
    }
    
    if (status != 0)
      break;

    printf ("\nThe encrypted data (%u bytes):\n", encryptedData->len);
    PrintBuf (encryptedData->data, encryptedData->len);
  } while (0);

   if (status != 0)
    printf ("Encryption failed.  Status = %i\n", status);

   B_DestroyAlgorithmObject (&rsaEncryptor);

   return status;
}

/*  Supports RSA private-key decryption using OAEP.
 *  decryptedData.data must point to a block of memory large enough
 *  to hold the decrypted data.
 */
int RSADecrypt (privateKey, encryptedData, decryptedData)
B_KEY_OBJ *privateKey;
ITEM *encryptedData;
ITEM *decryptedData;
{
  B_ALGORITHM_METHOD *DECRYPTION_CHOOSER[] = {
    &AM_RSA_CRT_DECRYPT_BLIND,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };
  
  B_ALGORITHM_OBJ rsaDecryptor = (B_ALGORITHM_OBJ)NULL_PTR;

  unsigned char outputBuffer[BLOCK_SIZE];
  
  unsigned int encryptedDataLeft;
  unsigned int decryptedBlockLen, decryptedPayloadLen;
  unsigned int encryptedDataIndex;
  unsigned int status, outputLenUpdate, outputLenFinal, padBytes;

  do {
    printf ("   Decrypting with the RSA Private Key \n");
    printf ("   =================================== \n");
    
    /*  Step 1:  Creating an Algorithm Object */
    if ((status = B_CreateAlgorithmObject (&rsaDecryptor)) != 0)
      break;

    /*  Step 2:  Set the algorithm object to AI_SET_OAEP_RSAPrivate */
    if ((status = B_SetAlgorithmInfo
         (rsaDecryptor, AI_SET_OAEP_RSAPrivate, NULL_PTR)) != 0)
      break;

    /*  Step 3:  Init -- Use the private key associated with the public
                 key used to encrypt */
    if ((status = B_DecryptInit
         (rsaDecryptor, *privateKey, DECRYPTION_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /*  Step 4:  Update */
    /*  The encrypted data length should be a multiple of the
        block size.  */
    
    encryptedDataIndex = 0;
    decryptedPayloadLen = 0;
    encryptedDataLeft = encryptedData->len;
    decryptedData->len = 0;

    /*  Decrypt each block of encrypted data  */
    while (encryptedDataLeft > 0) {
      if ((status = B_DecryptUpdate
                    (rsaDecryptor, outputBuffer, &outputLenUpdate,
                     BLOCK_SIZE, encryptedData->data + encryptedDataIndex,
                     BLOCK_SIZE, NULL_PTR,
                     (A_SURRENDER_CTX *)NULL_PTR)) != 0)
        break;

      /*  Step 5:  Final */
      if ((status = B_DecryptFinal
                    (rsaDecryptor, outputBuffer + outputLenUpdate,
                     &outputLenFinal, BLOCK_SIZE - outputLenUpdate,
                     NULL_PTR, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
        break;

      decryptedPayloadLen = outputLenUpdate + outputLenFinal;
      encryptedDataLeft -= BLOCK_SIZE;
      encryptedDataIndex += BLOCK_SIZE;

      /*  Copy the decrypted data from outputBuffer to decryptedData->data.
          The BC byte is not part of the payload; when encrypting we used
          BC=0.  Strip any trailing zeros, which were pad bytes.  */
      if (outputBuffer[0] != 0) {
        status = BE_INPUT_DATA;
        break;
      }
      
      for (padBytes = 0, decryptedBlockLen = decryptedPayloadLen - 1;
           (decryptedBlockLen > 0) && 
             (outputBuffer[decryptedBlockLen] == 0);
           ++padBytes, --decryptedBlockLen);

      /* amount of 'real' data in the decrypted data is 
         decryptedBlockLen bytes */
       
      T_memcpy (decryptedData->data + decryptedData->len, 
                outputBuffer + 1, decryptedBlockLen);
      decryptedData->len += decryptedBlockLen;
    }

    if (status != 0)
      break;
    
    printf ("\nThe decrypted data (%u bytes): \n", decryptedData->len);
    PrintBuf (decryptedData->data, decryptedData->len);
  } while (0);

  if (status != 0)
    printf ("Decryption failed.  Status = %i\n", status);
  
  B_DestroyAlgorithmObject (&rsaDecryptor);

  return status;
}

/*   General Surrender Function   */
/*   ==========================   */
int GeneralSurrenderFunction (handle)
POINTER handle;
{
  char s[100];
  static time_t currentTime;
  time_t getTime;

  if ((int)*handle == 0) {
    getTime = time(NULL);
    strftime (s, 100, "%H:%M:%S on %A, %d %B %Y", localtime(&getTime));
    printf ("\n%s\n", s);
    printf ("Surrender function ...\n");
    *handle = 1;
    time (&currentTime);
  }
  else {
    time (&getTime);
    if (currentTime != getTime) {
      printf (" .");
      currentTime = getTime;
    }
  }
  return (0);
} /*  end GeneralSurrenderFunction  */


/*  This procedure will print out what's in the buffer.
 */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 7) == 7) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

  printf ("\n");
} /*  end PrintBuf  */
