/* Copyright (C) RSA Data Security, Inc. created 1994.

   This file is used to demonstrate how to interface to an
   RSA Data Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security,
   Inc. has no warranty, implied or otherwise, or liability
   for this demonstration file or any modified version.
 */

#include <stdio.h>
#include "aglobal.h"
#include "bsafe.h"
#include "bslite.h"
#include "filebsl.h"

static int ReadFile PROTO_LIST
  ((char *, unsigned char *, unsigned int *, unsigned int));
static int WriteFile PROTO_LIST
  ((char *, unsigned char *, unsigned int));

/*****************************************************************************
 *
 * BSL_MD5WithRSASignFile
 * ----------------------
 * Creates a RSA Digital Signature on a data file using the MD5 message
 * digest algorithm and the supplied private key.  If the value of the
 * password is NULL_PTR, the private key should be supplied in its
 * unprotected form.  The signature is then stored in the specified file.
 *
 *****************************************************************************/

int BSL_MD5WithRSASignFile
  (signatureOutFilename, dataInFilename, ppPrivateKey, ppPrivateKeyLen,
   password, passwordLen, pbeParams, pbeParamsLen)
char *signatureOutFilename;
char *dataInFilename;
unsigned char *ppPrivateKey;      /* input: buffer of private key bytes */
unsigned int ppPrivateKeyLen;     /* input: number of private key bytes */
unsigned char *password;          /* input: buffer of private key protection */
                                  /*        password bytes */
unsigned int passwordLen;         /* input: number of password bytes */
unsigned char *pbeParams;         /* input: buffer of password-based */
                                  /*        encryption algorithm parameters */
                                  /*        bytes */
unsigned int pbeParamsLen;        /* input: number of PBE params bytes */
{
  B_ALGORITHM_OBJ algorithmObject = NULL_PTR;
  int status;
  unsigned char inputBuffer[1024], signature[512];
  unsigned int inputLen, signatureLen;
  FILE *file = (FILE *) NULL_PTR;
  
  do {
    /* Open for reading the input file. */
    if ((file = fopen (dataInFilename, "rb")) == NULL)
      /* could not open file */
    return (BE_DATA);
  
    /* Initializate algorithm objects used to calculate signatures */
    if ((status = BSL_MD5WithRSASignInit
         (&algorithmObject, ppPrivateKey, ppPrivateKeyLen, password,
          passwordLen, pbeParams, pbeParamsLen)) != 0)
      break;

    /* Create signature for contents of inputBuffer */
    do  {
    /* Iteratively read and process the input data in order to produce
       signature.
     */
      inputLen = fread (inputBuffer, 1, 1024, file);
      if ((status = BSL_MD5WithRSASignUpdate
        (&algorithmObject, inputBuffer, inputLen)) != 0)
        break;
    } while (inputLen == 1024); 
    if (status != 0)
      break;
    /* Finalizes the signature computation */
    if ((status = BSL_MD5WithRSASignFinal
         (&algorithmObject, signature, &signatureLen)) != 0)
      break;
    fclose (file);
    
    /* Write the signature to the specified file */
    if ((status = WriteFile
         (signatureOutFilename, signature, signatureLen)) != 0)
      break;
  } while (0);

  return (status);
}        
      
/*****************************************************************************
 *
 * BSL_MD5WithRSAVerifyFile
 * ------------------------
 * Verifies a RSA Digital Signature on a data file using the MD5 message
 * digest algorithm and the supplied public key.  A zero is returned if
 * the signature is successfully verified, a non-zero error value is
 * returned otherwise.
 *
 *****************************************************************************/

int BSL_MD5WithRSAVerifyFile
  (signatureInFilename, dataInFilename, publicKey, publicKeyLen)
char *signatureInFilename;
char *dataInFilename;
unsigned char *publicKey;  /* input: buffer of public key bytes */
unsigned int publicKeyLen; /* input: number of public key bytes */
{
  B_ALGORITHM_OBJ algorithmObject = NULL_PTR;
  int status;
  unsigned char inputBuffer[1024], signature[512];
  unsigned int inputLen, signatureLen;
  FILE *file = (FILE *) NULL_PTR;
  
  do {
    /* Open for reading file to be verified */
    if ((file = fopen (dataInFilename, "rb")) == NULL)
      /* could not open file */
    return (BE_DATA);
      /* Read the signature that is used to verify the inputBuffer */
    if ((status = ReadFile
          (signatureInFilename, signature, &signatureLen, 512)) != 0)
      break;
    /* Initializate algorithm objects used to verify signatures */
    if ((status = BSL_MD5WithRSAVerifyInit
         (&algorithmObject, publicKey, publicKeyLen)) != 0)
        break;
    /* Iteratively read and process data in input file in order to verify its
       signature.
     */
    do  {
      inputLen = fread (inputBuffer, 1, 1024, file);
      if ((status = BSL_MD5WithRSAVerifyUpdate
         (&algorithmObject, inputBuffer, inputLen)) != 0)
        break;
    } while (inputLen == 1024);
    /* Finalizes signature verification using the supplied signature */
    if ((status = BSL_MD5WithRSAVerifyFinal
         (&algorithmObject, signature, signatureLen)) != 0)
        break;
    fclose (file);
  } while (0);

  /* A zero return status means signature verification was successful */
  return (status);
}

/*****************************************************************************
 *
 * BSL_SealFileDESWithRSAEnvelope
 * ------------------------------
 * Computes an RSA Digital Envelope for a data file using a random key
 * for DES data encryption.  The DES key is then encrypted with the supplied
 * RSA public key.  The DES key and IV values are both generated by an
 * random object using the supplied seed value.
 *
 *****************************************************************************/

int BSL_SealFileDESWithRSAEnvelope
  (encryptedKeyOutFilename, ivOutFilename, encryptedDataOutFilename,
   dataInFilename, seed, seedLen, publicKey, publicKeyLen)
char *encryptedKeyOutFilename;
char *ivOutFilename;
char *encryptedDataOutFilename;
char *dataInFilename;
unsigned char *seed;           /* input: buffer of seed bytes */
unsigned int seedLen;          /* input: number of seed bytes */
unsigned char *publicKey;      /* input: buffer of public key bytes */
unsigned int publicKeyLen;     /* input: number of public key bytes */
{
  B_ALGORITHM_OBJ algorithmObject = NULL_PTR;
  int status;
  unsigned char inputData[1024], encryptedData[1024],
    encryptedKey[1024], IV[8];
  unsigned int inputDataLen, encryptedKeyLen, encryptedDataLen, len;
  FILE *inputFile  = (FILE *) NULL_PTR;
  FILE *outputFile  = (FILE *) NULL_PTR;
  
  do {
    /* Open for reading plaintext file */
    if ((inputFile = fopen (dataInFilename, "rb")) == NULL) {
      /* could not open file */
      status = BE_DATA;
      break;
    }
    /* Open output file for writing */
   if (encryptedDataOutFilename[0] == '-'
       && encryptedDataOutFilename[1] == '\0')
    /* use stdout */
    outputFile = stdout;
   else 
     if ((outputFile = fopen (encryptedDataOutFilename, "wb")) == NULL) {
      /* could not open file */
       status = BE_DATA;
       break;
    } 
    /* Sets up algorithm object for computing the RSA envelope */
    if ((status = BSL_SealDESWithRSAEnvelopeSetup
         (&algorithmObject, encryptedKey, &encryptedKeyLen, IV, seed,
          seedLen, publicKey, publicKeyLen)) != 0)
      break;

    /* Do the actual digital envelope encryption by iteratively reading data
       in chunks of 1024 bytes.
     */
    do {
      inputDataLen = fread (inputData, 1, 1024, inputFile);
      if ((status = BSL_BlockEncrypt
          (&algorithmObject, encryptedData, &encryptedDataLen, inputData,
           inputDataLen)) != 0)
        break;
      if (fwrite
        (encryptedData, 1, encryptedDataLen, outputFile) < encryptedDataLen) {
        status = BE_DATA;
        break;
      }
    } while (inputDataLen == 1024);
    if (status != 0)
      break;
    /* Pass in NULL_PTR for the input to indicate the end of input */
    if ((status = BSL_BlockEncrypt
         (&algorithmObject, encryptedData, &len,
          (unsigned char *)NULL_PTR, 0)) != 0)
      break;
    /* Write out last of encrypted data to file. */
    if (fwrite (encryptedData, 1, len, outputFile) < len) {
      status = BE_DATA;
      break;
      }
    
    /* Write the IV to the specified file */
    if ((status = WriteFile
         (ivOutFilename, IV, 8)) != 0)
      break;

    /* Write the encrypted DES key to the specified file */
    if ((status = WriteFile
         (encryptedKeyOutFilename, encryptedKey, encryptedKeyLen)) != 0)
      break;

  } while (0);
  if (inputFile != NULL)
    fclose (inputFile);
  if (outputFile != NULL && outputFile != stdout)
    fclose (outputFile);
  return (status);
}

/*****************************************************************************
 *
 * BSL_OpenFileDESWithRSAEnvelope
 * ------------------------------
 * Opens an RSA Digital Envelope for a data file using the supplied private
 * key.  If the password is NULL_PTR, the private key must be in its
 * unprotected form.
 *
 *****************************************************************************/

int BSL_OpenFileDESWithRSAEnvelope
  (dataOutFilename, encryptedKeyInFilename, ivInFilename,
   encryptedDataInFilename, ppPrivateKey, ppPrivateKeyLen, password,
   passwordLen, pbeParams, pbeParamsLen)
char *dataOutFilename;
char *encryptedKeyInFilename;
char *ivInFilename;
char *encryptedDataInFilename;
unsigned char *ppPrivateKey;  /* input: buffer of private key bytes */
unsigned int ppPrivateKeyLen; /* input: number of private key bytes */
unsigned char *password;      /* input: buffer of password bytes */
unsigned int passwordLen;     /* input: number of password bytes */
unsigned char *pbeParams;     /* input: buffer of password-based encryption */
                              /*        algorithm parameters bytes */
unsigned int pbeParamsLen;    /* input: number of PBE params bytes */
{
  B_ALGORITHM_OBJ algorithmObject = NULL_PTR;
  int status;
  unsigned char outputData[1024], encryptedData[1024],
    encryptedKey[1024], IV[8];
  unsigned int outputDataLen, encryptedKeyLen, encryptedDataLen, len, ivLen;
  FILE *outputFile = (FILE *) NULL_PTR;
  FILE *inputFile = (FILE *) NULL_PTR;
  do {
    /* Open the file of encrypted data */
    if ((inputFile = fopen (encryptedDataInFilename, "rb")) == NULL) {
      /* could not open file */
      status = BE_DATA;
      break;
    }
    /* Open the output file for writing */
    if (dataOutFilename[0] == '-'
       && dataOutFilename[1] == '\0')
      /* use stdout */
      outputFile = stdout;
    else 
      if ((outputFile = fopen (dataOutFilename, "wb")) == NULL) {
      /* could not open file */
        status = BE_DATA;
        break;
    } 

    /* Read in the encrypted DES key that is to be used to open the envelope */
    if ((status = ReadFile
         (encryptedKeyInFilename, encryptedKey, &encryptedKeyLen,
          1024)) != 0)
      break;

    /* Read in the IV */
    if ((status = ReadFile
         (ivInFilename, IV, &ivLen, 8)) != 0)
      break;


    /* Set up algorithm Object to open an RSA envelope */
    if ((status = BSL_OpenDESWithRSAEnvelopeSetup
         (&algorithmObject, encryptedKey, encryptedKeyLen, IV, ppPrivateKey,
          ppPrivateKeyLen, password, passwordLen, pbeParams,
          pbeParamsLen)) != 0)
      break;

    /* Perform the actual digital envelope decryption by iteratively
       reading through encrypted data in chunks of 1024 bytes.
     */
    do {
      encryptedDataLen = fread (encryptedData, 1, 1024, inputFile);
      if ((status = BSL_BlockDecrypt
         (&algorithmObject, outputData, &outputDataLen, encryptedData,
          encryptedDataLen)) != 0)
        break;
      if (fwrite
        (outputData, 1, outputDataLen, outputFile) < outputDataLen) {
        status = BE_DATA;
        break;      
      }
    } while (encryptedDataLen == 1024);
    if (status != 0)
      break;
    /* Pass in NULL_PTR to indicate end of input data stream */
    if ((status = BSL_BlockDecrypt
         (&algorithmObject, outputData, &len,
          (unsigned char *)NULL_PTR, 0)) != 0)
      break;

    /* Write the last data to a file */
     if (fwrite (outputData, 1, len, outputFile) < len) {
      status = BE_DATA;
      break;
      }
    
  } while (0);
  if (inputFile != NULL)
    fclose (inputFile);
  if (outputFile != NULL && outputFile != stdout)
    fclose (outputFile);
  return (status);
}
    
/* Read a file of up to length maxOutputLen bytes, storing it in
     output and returning its length in outputLen.
   Return 0 on success or 1 if error or if user cancels by entering a blank.
 */
static int ReadFile (filename, output, outputLen, maxOutputLen) 
char *filename;
unsigned char *output;
unsigned int *outputLen;
unsigned int maxOutputLen;
{
  FILE *file = (FILE *) NULL_PTR;
  unsigned char dummy;
  
  if ((file = fopen (filename, "rb")) == NULL)
      /* could not open file */
    return (BE_DATA);
  
  /* fread () returns the number of items read in.  Expect an end of file
       after the read.
   */
  *outputLen = fread (output, 1, maxOutputLen, file);
  if (*outputLen == maxOutputLen)
    /* Read exactly maxOutputLen bytes, so reading one more will set 
         end of file if there were exactly maxOutputLen bytes in the file.
     */
    fread (&dummy, 1, 1, file);
  
  if (!feof (file)) {
    fclose (file);
    return (BE_DATA);
  }
  
  fclose (file);
  return (0);
}

/* Write the input data of length inputLen to a file.
   Return 0 on success or 1 if error or if user cancels by entering a blank.
 */
static int WriteFile (filename, input, inputLen) 
char *filename;
unsigned char *input;
unsigned int inputLen;
{
  FILE *file = (FILE *) NULL_PTR;
  
  if (filename[0] == '-' && filename[1] == '\0')
    /* use stdout */
    file = stdout;
  else {
    if ((file = fopen (filename, "wb")) == NULL)
      /* could not open file */
      return (BE_DATA);
  }
  
  if (fwrite (input, 1, inputLen, file) < inputLen) {
    fclose (file);
    return (BE_DATA);
  }
  else {
    if (file == stdout)
      /* Printing to screen, so print a new line. */
      printf ("\n");
  }

  if (file != stdout)
    fclose (file);
  
  return (0);
}

