/* Copyright (C) RSA Data Security, Inc. created 1999.

   This file is used to demonstrate how to interface to an RSA Data
   Security, Inc. licensed development product.

   You have a royalty-free right to use, modify, reproduce and
   distribute this demonstration file (including any modified
   version), provided that you agree that RSA Data Security, Inc. has
   no warranty, implied or otherwise, or liability for this
   demonstration file or any modified version. 
*/

/* This program will generate a series of random values using the
   Intel hardware random device. */

#include <stdio.h>
#include <string.h>

#include "bsafe.h"

#define BYTES_TO_GENERATE 128

void PrintBuf PROTO_LIST((unsigned char *, unsigned int));

/* GenerateSeed will place a random seed in seedBytes.  The amount of
   data placed in seedBytes will be less than or equal to
   seedMaxLength.  If the Intel hardware random number generator is
   not available, then GenerateSeed will fall back to gathering seed
   data from the user. */
int GenerateSeed (unsigned char *seedBytes, unsigned int seedMaxLength) {

  B_ALGORITHM_OBJ randomAlgorithm = (B_ALGORITHM_OBJ)NULL_PTR;

  POINTER *oemTagList = (POINTER *)NULL_PTR;
  B_ALGORITHM_METHOD **CHOOSER = (B_ALGORITHM_METHOD **)NULL_PTR;

  /* The software chooser.  The AM_HW_RANDOM method in this chooser
     will be replaced with the hardware method. */
  B_ALGORITHM_METHOD *SOFTWARE_CHOOSER[] = {
    &AM_HW_RANDOM,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  /* This table lists the hardware algorithm methods that could be
     available.  This list isn't the list of algorithms that will be
     used.  That is determined by the hardware availablity. */
  HW_TABLE_ENTRY *HARDWARE_CHOOSER[] = {
    &HW_INTEL_RANDOM, 
    (HW_TABLE_ENTRY *)NULL_PTR
  };

  unsigned int status, index;
  unsigned int intelHardwarePresent = 0;

  do {

    /* Step 0: Create the session chooser.  This call will combine the
       elements of the software chooser, with those in the hardware
       chooser.  For each item in the software chooser that references
       a hardware method, the hardware method replaces it.

       This function will associate a hardware method with a software
       method so that when a software method is called, it turns to
       the hardware.  In this case, HW_INTEL_RANDOM will be called
       when AM_HW_RANDOM is referenced. 

       If more than one hardware method can be matched to a single
       software method, then the hardware method listed first in the
       hardware chooser is associated with that software method.  The
       list can be modified at run time by creating the hardware
       chooser at run time. */
    if ((status = B_CreateSessionChooser 
         (SOFTWARE_CHOOSER, &CHOOSER, (POINTER *)HARDWARE_CHOOSER,
          (ITEM *)NULL_PTR, NULL_PTR, &oemTagList)) != 0)
      break;
  
    printf ("[Seed]  Session chooser successfully created.\n");
    printf ("[Seed]  OEM tag(s) returned by B_CreateSessionChooser:\n");
    index = 0;
    do {
      status = strcmp (oemTagList[index], "AM_HW_RANDOM.com.Intel");
      if (status == 0)
        intelHardwarePresent = 1;
      printf ("     %s\n", oemTagList[index++]);
    } while (oemTagList[index] != NULL_PTR);
    printf ("[Seed]  End of OEM tag list.\n");
    if (intelHardwarePresent) {
      printf ("[Seed]  Intel hardware found.  Continuing.\n");
    } else {
      printf ("[Seed]  Intel hardware not found.  Using software.\n");
      status = 1; break;
    }

    /* Step 1: Create the algorithm object.  This object will
       "control" the random byte generation, storing any state that
       needs to be stored. */
    if ((status = B_CreateAlgorithmObject (&randomAlgorithm)) != 0)
      break;
    printf ("[Seed]  Created the random algorithm object.\n");

    /* Step 2: Set the algorithm info.  We will specify AI_HW_Random,
       which will point to the hardware method that was associated to
       AI_HW_Random in the call to B_CreateSessionChooser. */
    if ((status = B_SetAlgorithmInfo
         (randomAlgorithm, AI_HW_Random, NULL_PTR)) != 0)
      break;
    printf ("[Seed]  Set the algorithm info.\n");

    /* Step 3: Initialize randomAlgorithm to generate random bytes.
       Here we pass the CHOOSER that was created in the call to
       B_CreateSessionChooser above.  This chooser contains pointers
       to the hardware method that was associated with
       AI_HW_Random.  

       If the Intel hardware was not available, then we will fall back
       to gathering the seed from the user. */
    if ((status = B_RandomInit 
         (randomAlgorithm, CHOOSER, (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
    printf ("[Seed]  Initialized the random algorithm object for\n");
    printf ("        random number generation.\n");

    /* Step 5: Generate the random bytes.  Here we are asking the
       Crypto-C SDK to generate some 64 random bytes, storing the data
       in seedBytes.  The last parameter is a surrender context.  In
       this case, generating 64 random bytes should be very quick, so
       we will pass in a properly cast NULL_PTR. */
    if ((status = B_GenerateRandomBytes
         (randomAlgorithm, seedBytes, seedMaxLength, 
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    /* Step 6: Destroy the algorithm object.  This step will free any
       allocated memory used by randomAlgorithm.  The memory is
       zero'ed out before it is deallocated, so that potentially
       sensitive information is not left in memory. */
    B_DestroyAlgorithmObject (&randomAlgorithm);
    printf ("[Seed]  Destroyed random algorithm object.\n");

    /* Step 6b: Free the session chooser.  It is important to free the
       session chooser, so that any handles to hardware and allocated
       memory are released. */
    if ((status = B_FreeSessionChooser (&CHOOSER, &oemTagList)) != 0)
      break;
    printf ("[Seed]  Free'd session chooser.\n");
  } while (0);

  /* If an error has ocurred, then print out a summary of the error. */
  if (status != 0) {
    /* There's been an error.  Return the status code. */
    printf ("[Seed]  Error ocurred during random number generation.\n");
    printf ("[Seed]  Status = %d\n", status);
    
    /* Print out the extended error information, if available.  This
       information will give us a better idea of what went wrong with
       the hardware. */
    {
      ITEM errorData;
      POINTER am;
      
      /* Call B_GetExtendedErrorInfo to retrieve the error information.
         The data returned in the data field of errorData is a structure
         of A_RSA_EXTENDED_ERROR type.  The third parameter will contain
         a pointer to the alorithm method (AM) that was in use when the
         error was encountered. */
      B_GetExtendedErrorInfo(randomAlgorithm, &errorData, &am);
      
      if (am = &HW_INTEL_RANDOM) {
        /* Print out the error information. */
        printf ("[Seed]  Code: %d\n", 
                ((A_RSA_EXTENDED_ERROR *)errorData.data)->errorCode);
        printf ("[Seed]  Message: %s", 
                ((A_RSA_EXTENDED_ERROR *)errorData.data)->errorMsg);
      }
    }
    
    /* Try the following to deallocate and clear any memory properly. */
    B_DestroyAlgorithmObject (&randomAlgorithm);
    if ((CHOOSER != (B_ALGORITHM_METHOD **)NULL_PTR) || 
        (oemTagList != (POINTER *)NULL_PTR))
      B_FreeSessionChooser (&CHOOSER, &oemTagList);
    
    /* An error occurred, so it is necessary to request a seed
       directly from the user.  In this case, we will demonstrate the
       poor method of having the user type in the seed.  A better
       method would use the intra-keystroke timings along with mouse
       movement timings. */
    printf ("[Seed]  Please type a seed (%d chars max):\n", 
            seedMaxLength);
    gets (seedBytes);
  }

  printf ("[Seed]  Random seed:\n");
  PrintBuf (seedBytes, seedMaxLength);

  printf ("[Seed]  Finished executing.\n\n");
  return (seedMaxLength);
}

void main()
{
  B_ALGORITHM_OBJ randomAlgorithm = (B_ALGORITHM_OBJ)NULL_PTR;
  unsigned char *randomSeed = NULL_PTR;
  unsigned int randomSeedLen;
  unsigned char *randomByteBuffer = NULL_PTR;
  unsigned int status;

  B_ALGORITHM_METHOD *RANDOM_CHOOSER[] = {
    &AM_SHA_RANDOM,
    (B_ALGORITHM_METHOD *)NULL_PTR
  };

  do {
    printf ("[INTELRND]  Generating random bytes.\n");
    
    /* Step 1: Create a random algorithm object. */
    if ((status = B_CreateAlgorithmObject (&randomAlgorithm)) != 0)
      break;
    
    /* Step 2: Set the random algorithm object to SHA1. */
    if ((status = B_SetAlgorithmInfo
         (randomAlgorithm, AI_SHA1Random, NULL_PTR)) != 0)
      break;
    
    /* Step 3: Initialize the random algorithm. */
    if ((status = B_RandomInit
         (randomAlgorithm, RANDOM_CHOOSER,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
    
    /* Step 4: Get a random seed, and pass it to the random algorithm
       object.  This seed will be generated by the Intel hardware
       random device.  If the device cannot be found, or fails during
       processing, we will request that the user enters a seed
       value. */
    printf ("[INTELRND]  Getting seed value.\n");
    randomSeedLen = BYTES_TO_GENERATE;
    randomSeed = (unsigned char *)T_malloc (randomSeedLen);
    GenerateSeed (randomSeed, randomSeedLen);

    /* Step 4b: Now we have a random seed and its length.  Pass both
       into B_RandomUpdate. */
    if ((status = B_RandomUpdate
         (randomAlgorithm, randomSeed, randomSeedLen,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;
    
    /* Step 5: Generate.  First, prepare a buffer for receiving the
       random bytes before calling B_GenerateRandomBytes. */
    randomByteBuffer = T_malloc (BYTES_TO_GENERATE);
    if ((status = (randomByteBuffer == NULL_PTR)) != 0)
          break;
    printf ("[INTELRND]  Generated random bytes.\n");

    T_memset (randomByteBuffer, 0, BYTES_TO_GENERATE);

    if ((status = B_GenerateRandomBytes
         (randomAlgorithm, randomByteBuffer, BYTES_TO_GENERATE,
          (A_SURRENDER_CTX *)NULL_PTR)) != 0)
      break;

    printf ("[INTELRND]  Random data (%d bytes): \n", BYTES_TO_GENERATE);
    PrintBuf (randomByteBuffer, BYTES_TO_GENERATE);

  } while (0);

  if (status != 0) {
    printf ("[INTELRND]  Status = %i \n", status);      
    printf ("[INTELRND]  Generate random bytes failed.");
  }

  /*  Step 6:  Remember to destroy all objects, and free all memory
   */
  B_DestroyAlgorithmObject (&randomAlgorithm);
  T_memset (randomSeed, 0, randomSeedLen);
  T_free (randomSeed);
  T_free (randomByteBuffer);

} /*  end main  */


/*  This procedure will print out what's in a specified data buffer.  */
void PrintBuf (buffer, bufferLen)
unsigned char *buffer;
unsigned int bufferLen;
{
  unsigned int i;

  for (i = 0; i < bufferLen; ++i) {
    if ( ((i & 15) == 15) || (i == (bufferLen - 1)) )
      printf ("  %02x\n", buffer[i]);
    else
      printf ("  %02x", buffer[i]);
  }

} /*  end PrintBuf  */


/* This function is required to link properly when using the Intel
   hardware random device.  In this case, though, we will not use the
   function, so we will return a 0, to indicate success. */
int T_GetDynamicList (HW_TABLE_LIST *hardwareTable, POINTER *tagList) {
  /* These arguments are not used, so we will "use" them a bit to quiet
     the compiler. */
  UNUSED_ARG(hardwareTable);
  UNUSED_ARG(tagList);

  return 0;
}
